import abc
import logging
from typing import Any
from typing import Optional

import numpy as np

from substrafl import exceptions

logger = logging.getLogger(__name__)


class BaseIndexGenerator(abc.ABC):
    """Base class for the index generator, must be
    subclassed.
    """

    def __init__(
        self,
        batch_size: Optional[int],
        num_updates: int,
        shuffle: bool = True,
        drop_last: bool = False,
        seed: int = 42,
    ):
        """
        Args:
            batch_size (typing.Optional[int]): The size of each batch. If set to None, the batch_size will be
                the number of samples.
            num_updates (int): Number of local updates at each round
            shuffle (bool, Optional): Shuffle the indexes or not. Defaults to True.
            drop_last (bool, Optional): Drop the last batch if its size is inferior to the batch size. Defaults to
                False.
            seed (int, Optional): Random seed. Defaults to 42.

        Raises:
            ValueError: if batch_size is negative
        """
        if batch_size is None:
            logger.info("None was passed as a batch size. It will be set to n_sample size.")
        elif batch_size < 0:
            raise ValueError(f"batch_size must be positive but {batch_size} was passed.")

        self._batch_size: Optional[int] = batch_size
        self._rng = np.random.default_rng(seed)
        self._shuffle: bool = shuffle
        self._drop_last: bool = drop_last
        self._num_updates: int = num_updates
        self._counter: int = 0
        self._n_epoch_generated: int = 0

        self._n_samples: Optional[int] = None

    @property
    def batch_size(self) -> int:
        """Number of samples used per batch.

        Returns:
            int: Batch size used by the index generator
        """
        return self._batch_size

    @property
    def counter(self) -> int:
        """Number of calls made to the iterator since the last counter reset.

        Returns:
            int: Number of calls made to the iterator
        """
        return self._counter

    @property
    def n_epoch_generated(self) -> int:
        """Number of epochs generated

        Returns:
            int: number of epochs generated
        """
        return self._n_epoch_generated

    @property
    def num_updates(self) -> int:
        """Number of batches generated between resets of the counter.

        Returns:
            int: number of updates
        """
        return self._num_updates

    @property
    def n_samples(self) -> Optional[int]:
        """Returns the number of samples in the dataset.

        Returns:
            typing.Optional[int]: number of samples in the dataset.
        """
        return self._n_samples

    @n_samples.setter
    def n_samples(self, _n_samples: int):
        """Set the number of samples in the dataset.

        The indexes returned at each batch are between 0 and ``_n_samples - 1``.

        Args:
            _n_samples (int): Number of samples in the dataset

        Raises:
            ValueError: if _n_samples is negative.
        """
        # Validates n_samples
        if _n_samples < 0:
            raise ValueError(f"n_samples must be non negative but {_n_samples} was passed.")

        self._n_samples: int = _n_samples

        # Validates batch_size
        if self._batch_size is None:
            logger.info("None was passed as a batch size. It is set to n_sample size.")
            self._batch_size: int = self._n_samples
        elif self._batch_size > _n_samples:
            logger.info(
                (
                    "The batch size ({batch_size}) is greater than the number of samples: n_samples."
                    "This is not allowed. Batch_size is now updated to equal number of samples ({n_samples})"
                ).format(batch_size=self._batch_size, n_samples=_n_samples)
            )
            self._batch_size = self._n_samples

    def __iter__(self) -> "BaseIndexGenerator":
        """Required methods for generators, returns ``self``."""
        return self

    @abc.abstractclassmethod
    def __next__(self) -> Any:
        """Shall return a python object (batch_index) which
        is used for selecting each batch in the training loop method during training in this way :
        `x[batch_index], y[batch_index]`

        Shall also update self._counter and self._n_epoch_generated

        After ``num_updates`` call, raises StopIteration.

        Returns:
            Any: The batch indexes.
        """
        raise NotImplementedError

    def reset_counter(self):
        """Reset the counter to
        prepare for the next generation
        of batches.
        """
        self._counter = 0

    def check_num_updates(self):
        """Check if the counter is equal to ``num_updates``, which means that ``num_updates``
        batches have been generated since this instance has been created or the counter has been reset.

        Raises:
            exceptions.IndexGeneratorUpdateError: if the counter is different from ``num_updates``.
        """
        if self.counter != self._num_updates:
            raise exceptions.IndexGeneratorUpdateError(
                "The batch index generator has not been updated properly, it was called"
                f" {self.counter} times against {self._num_updates} expected."
            )



import logging
from typing import Optional

import numpy as np

from substrafl import exceptions
from substrafl.index_generator.base import BaseIndexGenerator

logger = logging.getLogger(__name__)


class NpIndexGenerator(BaseIndexGenerator):
    """An index based batch generator. It returns an array of size ``batch_size`` indexes.
    If ``batch_size`` is equal to zero, this returns an empty array.

    Each batch is generated and returned via the method
    :py:func:`~substrafl.index_generator.np_index_generator.NpIndexGenerator.__next__`:

    .. code-block:: python

        batch_generator = NpIndexGenerator(batch_size=32, num_updates=100)
        batch_generator.n_samples = 10

        batch_1 = next(batch_generator)
        batch_2 = next(batch_generator)
        # ...
        batch_n = next(batch_generator)

    In that case, as the default seed is set, the results are deterministic:

    .. code-block:: python

            batch_1 = np.array([5, 6, 0])
            batch_12 = np.array([8, 4, 0])

    This class is stateful and can be saved and loaded with the pickle library:

    .. code-block:: python

        # Saving
        with open(indexer_path, "wb") as f:
            pickle.dump(batch_generator, f)
            f.close()

        # Loading
        with open(indexer_path, "rb") as f:
            loaded_batch_generator = pickle.load(f)
            f.close()

    This index generator can be used to generate the batches for one epoch. For that, ``num_updates`` must be equal to:

    If ``drop_last=True``:

    .. code-block:: python

        num_updates = math.floor(num_samples / batch_size)

    If ``drop_last=False``:

    .. code-block:: python

        num_updates = math.ceil(num_samples / batch_size)

    """

    def __init__(
        self,
        batch_size: Optional[int],
        num_updates: int,
        shuffle: bool = True,
        drop_last: bool = False,
        seed: int = 42,
    ):
        """
        Args:
            batch_size (typing.Optional[int]): The size of each batch. If set to None, the batch_size is the
                number of samples.
            num_updates (int): The number of updates. After num_updates, the generator raises a StopIteration error.
                To reset it for the next round, use the
                :py:func:`~substrafl.index_generator.np_index_generator.NpIndexGenerator.reset_counter` function.
            shuffle (bool, Optional): Set to True to shuffle the indexes before each new epoch. Defaults to True.
            drop_last (bool, Optional): Set to True to drop the last incomplete batch, if the dataset size is not
                divisible by the batch size. If False and the size of dataset is not divisible by the batch size, then
                the last batch is smaller. Defaults to False.
            seed (int, Optional): The seed to set the randomness of the generator and have reproducible results.
                Defaults to 42.
        """

        # Initialization
        super().__init__(
            batch_size=batch_size,
            num_updates=num_updates,
            shuffle=shuffle,
            drop_last=drop_last,
            seed=seed,
        )

    def __iter__(self):
        """Required methods for generators, returns ``self``."""
        return self

    def __next__(self):
        """Generates the next batch.

        At the start of each iteration through the whole dataset, if ``shuffle`` is True then all the indices are
        shuffled.
        If there are less elements left in the dataset than ``batch_size``, then if ``drop_last`` is False, a batch
        containing the remaining elements is returned, else the last batch is dropped and the batch is created from the
        whole dataset.
        Each calls updates the ``counter`` by one, and each time it goes through an epoch, increases
        ``n_epoch_generated`` by one.

        Raises:
            StopIteration: when this function has been called ``num_updates`` times.

        Returns:
            numpy.ndarray: The batch indexes as a numpy array.
        """
        if self._n_samples is None:
            raise exceptions.IndexGeneratorSampleNoneError(
                "Please set the number of samples using the" "n_samples function before iterating through the batches."
            )

        if self._counter == self._num_updates:
            raise StopIteration

        batch, self._to_draw = np.split(self._to_draw, [self._batch_size])

        # If there are not enough indexes left for a complete round we re initialize
        if (self._drop_last and self._to_draw.shape[0] < self._batch_size) or (self._to_draw.shape[0] == 0):
            self._to_draw = np.arange(self._n_samples)

            # we shuffle if needed
            if self._shuffle:
                self._to_draw = self._rng.permutation(self._to_draw)

            self._n_epoch_generated += 1

        self._counter += 1
        return batch

    @BaseIndexGenerator.n_samples.setter
    def n_samples(self, _n_samples: int):
        """Set the number of samples to draw from, then initialize
        the indexes to draw from when generating the batches.

        Args:
            _n_samples (int): number of samples in the dataset.
        """
        super(NpIndexGenerator, self.__class__).n_samples.fset(self, _n_samples)
        self._n_batch_per_epoch: int = (
            (
                int(np.floor(self._n_samples / self._batch_size))
                if self._drop_last
                else int(np.ceil(self._n_samples / self._batch_size))
            )
            if self._batch_size != 0
            else 0
        )
        self._to_draw = np.arange(self._n_samples)
        if self._shuffle:
            self._to_draw = self._rng.permutation(self._to_draw)



"""
Decorators to wrap functions
so that they are executed on the remote organizations.
"""
from functools import wraps
from typing import Any
from typing import Callable
from typing import List
from typing import Optional

from substrafl.remote.operations import AggregateOperation
from substrafl.remote.operations import DataOperation
from substrafl.remote.remote_struct import RemoteStruct
from substrafl.remote.substratools_methods import RemoteMethod


def remote_data(method: Callable):
    """Decorator for a remote function containing a ``data_samples`` argument (e.g the ``Algo.train`` function)
    With this decorator, when the function is called, it is not executed but it returns a ``DataOperation``
    object containing all the informations needed to execute it later
    (see ``substrafl.remote.operations.DataOperation``).

        - The decorated function definition should have at least a shared_state argument
        -   If the decorated function is called without a ``_skip=True`` argument, the arguments required
            are the ones in ``remote_method_inner``, and it should have at least a ``data_samples`` argument
        -   If the decorated function is called with a ``_skip=True`` argument, it should have the arguments of its
            original definition
        - The decorated function should be within a class
        - The ``__init__`` of the class must be

            .. code-block:: python

                def __init__(self, *args, **kwargs):
                    self.args = args
                    self.kwargs = kwargs

        -   ``self.args`` and ``self.kwargs`` will be given to the init, any other init argument is ignored
            (not saved in the RemoteStruct)

    Args:
        method (Callable): Method to wrap so that it is executed on the remote server
    """

    @wraps(method)
    def remote_method_inner(
        self,
        data_samples: Optional[List[str]] = None,
        shared_state: Any = None,
        _skip: bool = False,
        _algo_name: Optional[str] = None,
        **method_parameters,
    ) -> DataOperation:
        """
        Args:
            data_samples (List[str]): The data samples paths. Defaults to None.
            shared_state (typing.Any): a shared state, could be a SharedStateRef object or anything else. Defaults to
                None.
            _skip (bool, Optional): if True, calls the decorated function. Defaults to False.
            _algo_name(str, Optional): opportunity to set a custom algo name.
                Default to None.

        Returns:
            DataOperation: resulting DataOperation
        """
        if _skip:
            return method(self=self, shared_state=shared_state, **method_parameters)

        assert data_samples is not None

        assert "datasamples" not in method_parameters.keys()

        return DataOperation(
            RemoteStruct(
                cls=self.__class__,
                cls_args=self.args,
                cls_kwargs=self.kwargs,
                method_name=method.__name__,
                method_parameters=method_parameters,
                algo_name=_algo_name,
                remote_cls=RemoteMethod,
            ),
            data_samples,
            shared_state,
        )

    return remote_method_inner


def remote(method: Callable):
    """Decorator for a remote function.
    With this decorator, when the function is called, it is not executed but it returns a ``AggregateOperation``
    object containing all the informations needed to execute it later (see
    ``substrafl.remote.operations.AggregateOperation``).

        - The decorated function definition should have at least a shared_state argument
        -   If the decorated function is called without a ``_skip=True`` argument, the arguments required
            are the ones in ``remote_method_inner``
        -   If the decorated function is called with a ``_skip=True`` argument, it should have the arguments of its
            original definition
        - The decorated function should be within a class
        - The ``__init__`` of the class must be

            .. code-block:: python

                def __init__(self, *args, **kwargs):
                    self.args = args
                    self.kwargs = kwargs

        -   ``self.args`` and ``self.kwargs`` will be given to the init, any other init argument is ignored
            (not saved in the RemoteStruct)

    Args:
        method (Callable): Method to wrap so that it is executed on the remote server
    """

    @wraps(method)
    def remote_method_inner(
        self,
        shared_states: Optional[List] = None,
        _skip: bool = False,
        _algo_name: Optional[str] = None,
        **method_parameters,
    ) -> AggregateOperation:
        if _skip:
            return method(self=self, shared_states=shared_states, **method_parameters)

        return AggregateOperation(
            RemoteStruct(
                cls=self.__class__,
                cls_args=self.args,
                cls_kwargs=self.kwargs,
                method_name=method.__name__,
                method_parameters=method_parameters,
                algo_name=_algo_name,
                remote_cls=RemoteMethod,
            ),
            shared_states,
        )

    return remote_method_inner



from pathlib import Path
from pathlib import PosixPath
from typing import List

from pydantic import BaseModel
from pydantic import Field
from pydantic import validator

from substrafl.exceptions import InvalidPathError


class Dependency(BaseModel):
    """Dependency pydantic class.


    .. note:: If you are using your current package as a local dependencies, be aware that folders named
        **local-worker** or with **tmp_substrafl** as prefix are ignored during the installation.

    Args:
        editable_mode (bool): If set to False, substra, substrafl and substratools used in the
            Dockerfiles submitted to Substra platform will be taken from pypi.
            If set to True, it will be the one installed in editable mode from your python environment.
            Defaults to False.
        dependencies (List[str]): Python packages installable from pypi.
        local_dependencies (List[pathlib.Path]): Local installable packages. The command
            `pip install -e .` will be executed in each of those folders hence a `setup.py` must be present in each
            folder.
        local_code (List[pathlib.Path]): Local relative imports used by your script. All files / folders must be
            at the same level than your script.
    """

    editable_mode: bool = False
    pypi_dependencies: List[str] = Field(default_factory=list)
    local_dependencies: List[PosixPath] = Field(default_factory=list)
    local_code: List[PosixPath] = Field(default_factory=list)

    @validator("local_dependencies", "local_code")
    def resolve_path(cls, v):  # noqa: N805
        """Resolve list of local code paths and check if they exist."""
        not_existing_paths = list()
        resolved_paths = list()
        for path in v:
            if not Path(path).exists():
                not_existing_paths += [f"\n\t{Path(path)} AS {Path(path).resolve()}"]
            else:
                resolved_paths.append(Path(path).resolve())

        if not_existing_paths:
            raise InvalidPathError(
                f"Couldn't resolve :{''.join(not_existing_paths)}\nPlease explicit the input path(s)."
            )

        return resolved_paths

    @validator("local_dependencies")
    def check_setup(cls, v):  # noqa: N805
        """Check the presence of a setup.py file in the provided paths."""
        not_installable = list()
        for dependency_path in v:
            installable_dir = (dependency_path / "setup.py").is_file() or (dependency_path / "pyproject.toml").is_file()
            if dependency_path.is_dir() and not installable_dir:
                not_installable.append(dependency_path)
            # The user can also give tar.gz archives
        if not_installable:
            raise InvalidPathError(
                f"Passed folder must be installable python package. "
                f"But neither setup.py or pyproject.toml was found in :{', '.join([str(p) for p in not_installable])}"
            )
        return v



import logging

from substrafl.__version__ import __version__  # noqa
from substrafl.evaluation_strategy import EvaluationStrategy
from substrafl.experiment import execute_experiment
from substrafl.index_generator.np_index_generator import NpIndexGenerator
from substrafl.logger import set_logging_level
from substrafl.model_loading import download_algo_files
from substrafl.model_loading import load_algo
from substrafl.schemas import StrategyName

set_logging_level(loglevel=logging.INFO)

__all__ = [
    "execute_experiment",
    "set_logging_level",
    "NpIndexGenerator",
    "EvaluationStrategy",
    "StrategyName",
    "load_algo",
    "download_algo_files",
]



import logging
from typing import Any
from typing import List
from typing import Optional

import torch

from substrafl.algorithms.pytorch import weight_manager
from substrafl.algorithms.pytorch.torch_base_algo import TorchAlgo
from substrafl.index_generator import BaseIndexGenerator
from substrafl.remote import remote_data
from substrafl.schemas import FedAvgAveragedState
from substrafl.schemas import FedAvgSharedState
from substrafl.schemas import StrategyName

logger = logging.getLogger(__name__)


class TorchFedAvgAlgo(TorchAlgo):
    """To be inherited. Wraps the necessary operation so a torch model can be trained in the Federated Averaging
    strategy.

    The ``train`` method:

        - updates the weights of the model with the aggregated weights,
        - initializes or loads the index generator,
        - calls the :py:func:`~substrafl.algorithms.pytorch.TorchFedAvgAlgo._local_train` method to do the local
          training
        - then gets the weight updates from the models and sends them to the aggregator.

    The ``predict`` method generates the predictions.

    The child class can override the :py:func:`~substrafl.algorithms.pytorch.TorchFedAvgAlgo._local_train` and
    :py:func:`~substrafl.algorithms.pytorch.TorchFedAvgAlgo._local_predict` methods, or other methods if necessary.

    To add a custom parameter to the ``__init__`` of the class, also add it to the call to ``super().__init__``
    as shown in the example with ``my_custom_extra_parameter``. Only primitive types (str, int, ...) are supported
    for extra parameters.

    Example:

        .. code-block:: python

            class MyAlgo(TorchFedAvgAlgo):
                def __init__(
                    self,
                    my_custom_extra_parameter,
                ):
                    super().__init__(
                        model=perceptron,
                        criterion=torch.nn.MSELoss(),
                        optimizer=optimizer,
                        index_generator=NpIndexGenerator(
                            num_updates=100,
                            batch_size=32,
                        ),
                        dataset=MyDataset,
                        my_custom_extra_parameter=my_custom_extra_parameter,
                    )
                def _local_train(
                    self,
                    train_dataset: torch.utils.data.Dataset,
                ):

                    # Create torch dataloader from the automatically instantiated dataset
                    # ``train_dataset = self._dataset(datasamples=datasamples, is_inference=False)`` is executed prior
                    #  the execution of this function
                    train_data_loader = torch.utils.data.DataLoader(train_dataset, batch_sampler=self._index_generator)

                    for x_batch, y_batch in train_data_loader:

                        # Forward pass
                        y_pred = self._model(x_batch)

                        # Compute Loss
                        loss = self._criterion(y_pred, y_batch)
                        self._optimizer.zero_grad()
                        loss.backward()
                        self._optimizer.step()

                        if self._scheduler is not None:
                            self._scheduler.step()
    """

    def __init__(
        self,
        model: torch.nn.Module,
        criterion: torch.nn.modules.loss._Loss,
        optimizer: torch.optim.Optimizer,
        index_generator: BaseIndexGenerator,
        dataset: torch.utils.data.Dataset,
        scheduler: Optional[torch.optim.lr_scheduler._LRScheduler] = None,
        with_batch_norm_parameters: bool = False,
        seed: Optional[int] = None,
        use_gpu: bool = True,
        *args,
        **kwargs,
    ):
        """
        The ``__init__`` functions is called at each call of the `train()` or `predict()` function
        For round>=2, some attributes will then be overwritten by their previous states in the `load()` function,
        before the `train()` or `predict()` function is ran.

        Args:
            model (torch.nn.modules.module.Module): A torch model.
            criterion (torch.nn.modules.loss._Loss): A torch criterion (loss).
            optimizer (torch.optim.Optimizer): A torch optimizer linked to the model.
            index_generator (BaseIndexGenerator): a stateful index generator.
                Must inherit from BaseIndexGenerator. The __next__ method shall return a python object (batch_index)
                which is used for selecting each batch from the output of the get_data method of the opener
                during training in this way: ``x[batch_index], y[batch_index]``.
                If overridden, the generator class must be defined either as part of a package or in a different file
                than the one from which the ``execute_experiment`` function is called.
                This generator is used as stateful ``batch_sampler`` of the data loader created from the given
                ``dataset``
            dataset (torch.utils.data.Dataset): an instantiable dataset class whose ``__init__`` arguments are
                ``x``, ``y`` and ``is_inference``. The torch datasets used for both training and inference will be
                instantiate from it prior to the ``_local_train`` execution and within the ``predict`` method.
                The ``__getitem__`` methods of those generated datasets must return both ``x`` (training data) and y
                (target values) when ``is_inference`` is set to ``False`` and only ``x`` (testing data) when
                ``is_inference`` is set to True.
                This behavior can be changed by re-writing the `_local_train` or `predict` methods.
            scheduler (torch.optim.lr_scheduler._LRScheduler, Optional): A torch scheduler that will be called at every
                batch. If None, no scheduler will be used. Defaults to None.
            with_batch_norm_parameters (bool): Whether to include the batch norm layer parameters in the fed avg
                strategy. Defaults to False.
            seed (typing.Optional[int]): Seed set at the algo initialization on each organization. Defaults to None.
            use_gpu (bool): Whether to use the GPUs if they are available. Defaults to True.
        """
        super().__init__(
            model=model,
            criterion=criterion,
            optimizer=optimizer,
            index_generator=index_generator,
            dataset=dataset,
            scheduler=scheduler,
            seed=seed,
            use_gpu=use_gpu,
            *args,
            **kwargs,
        )
        self._with_batch_norm_parameters = with_batch_norm_parameters

    @property
    def strategies(self) -> List[StrategyName]:
        """List of compatible strategies

        Returns:
            typing.List: typing.List[StrategyName]
        """
        return [StrategyName.FEDERATED_AVERAGING]

    @remote_data
    def train(
        self,
        datasamples: Any,
        shared_state: Optional[FedAvgAveragedState] = None,  # Set to None per default for clarity reason as
        # the decorator will do it if the arg shared_state is not passed.
    ) -> FedAvgSharedState:
        """Train method of the fed avg strategy implemented with torch. This method will execute the following
        operations:

            * instantiates the provided (or default) batch indexer
            * if a shared state is passed, set the parameters of the model to the provided shared state
            * train the model for n_updates
            * compute the weight update

        Args:
            datasamples (typing.Any): Input data returned by the ``get_data`` method from the opener.
            shared_state (FedAvgAveragedState, Optional): Dict containing torch parameters that
                will be set to the model. Defaults to None.

        Returns:
            FedAvgSharedState: weight update (delta between fine-tuned
            weights and previous weights)
        """

        # Create torch dataset
        train_dataset = self._dataset(datasamples, is_inference=False)

        if shared_state is None:
            # Instantiate the index_generator
            assert self._index_generator.n_samples is None
            self._index_generator.n_samples = len(train_dataset)
        else:
            assert self._index_generator.n_samples is not None
            # The shared states is the average of the model parameter updates for all organizations
            # Hence we need to add it to the previous local state parameters
            parameter_updates = [torch.from_numpy(x).to(self._device) for x in shared_state.avg_parameters_update]
            weight_manager.increment_parameters(
                model=self._model,
                updates=parameter_updates,
                with_batch_norm_parameters=self._with_batch_norm_parameters,
            )

        self._index_generator.reset_counter()

        old_parameters = weight_manager.get_parameters(
            model=self._model, with_batch_norm_parameters=self._with_batch_norm_parameters
        )

        # Train mode for torch model
        self._model.train()

        # Train the model
        self._local_train(train_dataset)

        self._index_generator.check_num_updates()

        self._model.eval()

        parameters_update = weight_manager.subtract_parameters(
            parameters=weight_manager.get_parameters(
                model=self._model,
                with_batch_norm_parameters=self._with_batch_norm_parameters,
            ),
            parameters_to_subtract=old_parameters,
        )

        # Re set to the previous state
        weight_manager.set_parameters(
            model=self._model,
            parameters=old_parameters,
            with_batch_norm_parameters=self._with_batch_norm_parameters,
        )

        return FedAvgSharedState(
            n_samples=len(train_dataset),
            parameters_update=[p.cpu().detach().numpy() for p in parameters_update],
        )

    def summary(self):
        """Summary of the class to be exposed in the experiment summary file

        Returns:
            dict: a json-serializable dict with the attributes the user wants to store
        """
        summary = super().summary()
        summary.update(
            {
                "with_batch_norm_parameters": str(self._with_batch_norm_parameters),
            }
        )
        return summary



import copy
import datetime
import json
import logging
import uuid
from pathlib import Path
from platform import python_version
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import substra
import substratools

import substrafl
from substrafl import exceptions
from substrafl.algorithms.algo import Algo
from substrafl.dependency import Dependency
from substrafl.evaluation_strategy import EvaluationStrategy
from substrafl.exceptions import KeyMetadataError
from substrafl.exceptions import LenMetadataError
from substrafl.nodes.aggregation_node import AggregationNode
from substrafl.nodes.node import OperationKey
from substrafl.nodes.train_data_node import TrainDataNode
from substrafl.remote.remote_struct import RemoteStruct
from substrafl.strategies.strategy import Strategy

logger = logging.getLogger(__name__)


def _register_operations(
    client: substra.Client,
    train_data_nodes: List[TrainDataNode],
    aggregation_node: Optional[AggregationNode],
    evaluation_strategy: Optional[EvaluationStrategy],
    dependencies: Dependency,
) -> Tuple[List[dict], Dict[RemoteStruct, OperationKey]]:
    """Register the operations in Substra: define the algorithms we need and submit them

    Args:
        client (substra.Client): substra client
        train_data_nodes (typing.List[TrainDataNode]): list of train data nodes
        aggregation_node (typing.Optional[AggregationNode]): the aggregation node for
            centralized strategies
        evaluation_strategy (typing.Optional[EvaluationStrategy]): the evaluation strategy
            if there is one dependencies
        (Dependency): dependencies of the train algo

    Returns:
        typing.Tuple[typing.List[dict], typing.Dict[RemoteStruct, OperationKey]]:
        tasks, operation_cache
    """
    # `register_operations` methods from the different organizations store the id of the already registered
    # algorithm so we don't add them twice
    operation_cache = dict()
    predict_algo_cache = dict()
    tasks = list()

    train_data_organizations_id = {train_data_node.organization_id for train_data_node in train_data_nodes}
    aggregation_organization_id = {aggregation_node.organization_id} if aggregation_node is not None else set()
    test_data_organizations_ids = (
        evaluation_strategy.test_data_nodes_org_ids if evaluation_strategy is not None else set()
    )

    authorized_ids = list(train_data_organizations_id | aggregation_organization_id | test_data_organizations_ids)
    permissions = substra.sdk.schemas.Permissions(public=False, authorized_ids=authorized_ids)

    for train_data_node in train_data_nodes:
        operation_cache = train_data_node.register_operations(
            client,
            permissions,
            cache=operation_cache,
            dependencies=dependencies,
        )

        tasks += train_data_node.tasks

    if evaluation_strategy is not None:
        for test_data_node in evaluation_strategy.test_data_nodes:
            predict_algo_cache = test_data_node.register_predict_operations(
                client,
                permissions,
                cache=predict_algo_cache,
                dependencies=dependencies,
            )

            tasks += test_data_node.predicttasks
            tasks += test_data_node.testtasks

    # The aggregation operation is defined in the strategy, its dependencies are
    # the strategy dependencies
    if aggregation_node is not None:
        operation_cache = aggregation_node.register_operations(
            client,
            permissions,
            cache=operation_cache,
            dependencies=Dependency(editable_mode=dependencies.editable_mode),
        )

        tasks += aggregation_node.tasks

    return tasks, operation_cache


def _save_experiment_summary(
    experiment_folder: Path,
    compute_plan_key: str,
    strategy: Strategy,
    num_rounds: int,
    algo: Algo,
    operation_cache: Dict[RemoteStruct, OperationKey],
    train_data_nodes: TrainDataNode,
    aggregation_node: Optional[AggregationNode],
    evaluation_strategy: EvaluationStrategy,
    timestamp: str,
    additional_metadata: Optional[Dict],
):
    """Saves the experiment summary in `experiment_folder`, with the name format `{timestamp}_{compute_plan.key}.json`

    Args:
        experiment_folder (typing.Union[str, pathlib.Path]): path to the folder where the experiment summary is saved.
        compute_plan_key (str): compute plan key
        strategy (substrafl.strategies.Strategy): strategy
        num_rounds (int): num_rounds
        algo (substrafl.algorithms.Algo): algo
        operation_cache (typing.Dict[RemoteStruct, OperationKey]): operation_cache
        train_data_nodes (TrainDataNode): train_data_nodes
        aggregation_node (typing.Optional[AggregationNode]): aggregation_node
        evaluation_strategy (EvaluationStrategy): evaluation_strategy
        timestamp (str): timestamp with "%Y_%m_%d_%H_%M_%S" format
        additional_metadata (dict, Optional): Optional dictionary of metadata to be shown on the Substra WebApp.
    """
    # create the experiment folder if it doesn't exist
    experiment_folder = Path(experiment_folder)
    experiment_folder.mkdir(exist_ok=True)
    experiment_summary = dict()

    # add attributes of interest and summaries of the classes to the experiment summary
    experiment_summary["compute_plan_key"] = compute_plan_key
    experiment_summary["strategy"] = type(strategy).__name__
    experiment_summary["num_rounds"] = num_rounds
    experiment_summary["algo"] = algo.summary()
    experiment_summary["algo_keys"] = {
        operation_cache[remote_struct]: remote_struct.summary() for remote_struct in operation_cache
    }
    experiment_summary["train_data_nodes"] = [train_data_node.summary() for train_data_node in train_data_nodes]
    experiment_summary["test_data_nodes"] = []
    if evaluation_strategy is not None:
        experiment_summary["test_data_nodes"] = [
            test_data_node.summary() for test_data_node in evaluation_strategy.test_data_nodes
        ]

    experiment_summary["aggregation_node"] = aggregation_node.summary() if aggregation_node is not None else None
    if additional_metadata is not None:
        experiment_summary["additional_metadata"] = additional_metadata

    # Save the experiment summary
    summary_file = experiment_folder / f"{timestamp}_{compute_plan_key}.json"
    summary_file.write_text(json.dumps(experiment_summary, indent=4))
    logger.info(("Experiment summary saved to {0}").format(summary_file))


def _check_evaluation_strategy(
    evaluation_strategy: EvaluationStrategy,
    num_rounds: int,
):
    if evaluation_strategy.num_rounds is None:
        evaluation_strategy.num_rounds = num_rounds
    elif evaluation_strategy.num_rounds != num_rounds:
        raise ValueError(
            "num_rounds set in evaluation_strategy does not match num_rounds set in the experiment: "
            f"{evaluation_strategy.num_rounds} is not {num_rounds}"
        )


def _check_additional_metadata(additional_metadata: Dict):
    unauthorized_keys = set(
        ("substrafl_version", "substra_version", "substratools_version", "python_version", "num_rounds")
    )
    invalid_keys = set(additional_metadata.keys()).intersection(unauthorized_keys)

    if len(invalid_keys) > 0:
        raise KeyMetadataError(
            f"None of: `{'`, `'.join(unauthorized_keys)}` can be used as"
            f" metadata key but `{' '.join(invalid_keys)}` were/was found"
        )

    for metadata_value in additional_metadata.values():
        if len(str(metadata_value)) > 100:
            raise LenMetadataError(
                "The maximum length of a value in the additional_metadata dictionary is 100 characters."
            )


def _get_packages_versions() -> dict:
    """Returns a dict containing substrafl, substra and substratools versions

    Returns:
        dict: substrafl, substra and substratools versions
    """

    return {
        "substrafl_version": substrafl.__version__,
        "substra_version": substra.__version__,
        "substratools_version": substratools.__version__,
        "python_version": python_version(),
    }


def execute_experiment(
    client: substra.Client,
    algo: Algo,
    strategy: Strategy,
    train_data_nodes: List[TrainDataNode],
    num_rounds: int,
    experiment_folder: Union[str, Path],
    aggregation_node: Optional[AggregationNode] = None,
    evaluation_strategy: Optional[EvaluationStrategy] = None,
    dependencies: Optional[Dependency] = None,
    clean_models: bool = True,
    name: Optional[str] = None,
    additional_metadata: Optional[Dict] = None,
    task_submission_batch_size: int = 500,
) -> substra.sdk.models.ComputePlan:
    """Run a complete experiment. This will train (on the `train_data_nodes`) and test (on the
    `test_data_nodes`) your `algo` with the specified `strategy` `n_rounds` times and return the
    compute plan object from the Substra platform.

    In substrafl, operations are linked to each other statically before being submitted to substra.

    The execution of:

        - the `self.perform_round` method from the passed strategy **num_rounds** times
        - the `self.predict` methods from the passed strategy

    generate the static graph of operations.

    Each element necessary for those operations (Tasks and Algorithms)
    is registered to the Substra platform thanks to the specified client.

    Finally, the compute plan is sent and executed.

    The experiment summary is saved in `experiment_folder`, with the name format `{timestamp}_{compute_plan.key}.json`

    Args:
        client (substra.Client): A substra client to interact with the Substra platform
        algo (Algo): The algorithm your strategy will execute (i.e. train and test on all the specified nodes)
        strategy (Strategy): The strategy by which your algorithm will be executed
        train_data_nodes (typing.List[TrainDataNode]): List of the nodes where training on data
            occurs evaluation_strategy (EvaluationStrategy, Optional): If None performance will not be measured at all.
            Otherwise measuring of performance will follow the EvaluationStrategy. Defaults to None.
        aggregation_node (typing.Optional[AggregationNode]): For centralized strategy, the aggregation
            node, where all the shared tasks occurs else None.
        num_rounds (int): The number of time your strategy will be executed
        dependencies (Dependency, Optional): Dependencies of the algorithm. It must be defined from
            the substrafl Dependency class. Defaults None.
        experiment_folder (typing.Union[str, pathlib.Path]): path to the folder where the experiment summary is saved.
        clean_models (bool): Clean the intermediary models on the Substra platform. Set it to False
            if you want to download or re-use intermediary models. This causes the disk space to fill
            quickly so should be set to True unless needed. Defaults to True.
        name (str, Optional): Optional name chosen by the user to identify the compute plan. If None,
            the compute plan name is set to the timestamp.
        additional_metadata(dict, typing.Optional): Optional dictionary of metadata to be passed to the Substra WebApp.
        task_submission_batch_size(int): The compute plan tasks are submitted by batch. The higher the batch size,
            the faster the submission, a batch size that is too high makes the submission fail.
            Rule of thumb: batch_size = math.floor(120000 / number_of_samples_per_task)

    Returns:
        ComputePlan: The generated compute plan
    """
    if dependencies is None:
        dependencies = Dependency()

    if strategy.name not in algo.strategies:
        raise exceptions.IncompatibleAlgoStrategyError(
            f"The algo {algo.__class__.__name__} is not compatible with the strategy {strategy.__class__.__name__},"
            f"named {strategy.name}. Check the algo strategies property: algo.strategies to see the list of compatible"
            "strategies."
        )

    train_data_nodes = copy.deepcopy(train_data_nodes)
    aggregation_node = copy.deepcopy(aggregation_node)
    strategy = copy.deepcopy(strategy)
    evaluation_strategy = copy.deepcopy(evaluation_strategy)

    train_organization_ids = [train_data_node.organization_id for train_data_node in train_data_nodes]

    if len(train_organization_ids) != len(set(train_organization_ids)):
        raise ValueError("Training multiple algorithms on the same organization is not supported right now.")

    if evaluation_strategy is not None:
        _check_evaluation_strategy(evaluation_strategy, num_rounds)
        # Reset the evaluation strategy
        evaluation_strategy.restart_rounds()

    cp_metadata = dict()
    if additional_metadata is not None:
        _check_additional_metadata(additional_metadata)
        cp_metadata.update(additional_metadata)

    # Adding substrafl, substratools and substra versions to the cp metadata
    cp_metadata.update(_get_packages_versions())

    # Adding rounds metadata to cp_meta_data
    cp_metadata["num_rounds"] = num_rounds

    logger.info("Building the compute plan.")

    additional_orgs_permissions = (
        evaluation_strategy.test_data_nodes_org_ids if evaluation_strategy is not None else set()
    )

    # create computation graph
    for round_idx in range(0, num_rounds + 1):
        strategy.perform_round(
            algo=algo,
            train_data_nodes=train_data_nodes,
            aggregation_node=aggregation_node,
            additional_orgs_permissions=additional_orgs_permissions,
            round_idx=round_idx,
            clean_models=clean_models,
        )

        if evaluation_strategy is not None and next(evaluation_strategy):
            strategy.predict(
                algo=algo,
                train_data_nodes=train_data_nodes,
                test_data_nodes=evaluation_strategy.test_data_nodes,
                round_idx=round_idx,
            )

    # Computation graph is created
    logger.info("Registering the algorithm to Substra.")
    tasks, operation_cache = _register_operations(
        client=client,
        train_data_nodes=train_data_nodes,
        aggregation_node=aggregation_node,
        evaluation_strategy=evaluation_strategy,
        dependencies=dependencies,
    )

    # Execute the compute plan
    logger.info("Registering the compute plan to Substra.")
    timestamp = str(datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S"))

    # Generate the compute plan key.
    compute_plan_key = str(uuid.uuid4())

    # save the experiment summary in experiment_folder
    _save_experiment_summary(
        experiment_folder=experiment_folder,
        compute_plan_key=compute_plan_key,
        strategy=strategy,
        num_rounds=num_rounds,
        algo=algo,
        operation_cache=operation_cache,
        train_data_nodes=train_data_nodes,
        aggregation_node=aggregation_node,
        evaluation_strategy=evaluation_strategy,
        timestamp=timestamp,
        additional_metadata=additional_metadata,
    )
    compute_plan = client.add_compute_plan(
        substra.sdk.schemas.ComputePlanSpec(
            key=compute_plan_key,
            tasks=tasks,
            name=name or timestamp,
            metadata=cp_metadata,
        ),
        auto_batching=True,
        batch_size=task_submission_batch_size,
    )
    logger.info(("The compute plan has been registered to Substra, its key is {0}.").format(compute_plan.key))
    return compute_plan



"""
Create the Substra algo assets and register them to the platform.
"""
import inspect
import logging
import os
import shutil
import tarfile
import tempfile
import typing
import warnings
from distutils import util
from pathlib import Path
from platform import python_version

import substra
import substratools
from packaging import version

import substrafl
from substrafl import exceptions
from substrafl.dependency import Dependency
from substrafl.nodes.node import InputIdentifiers
from substrafl.nodes.node import OutputIdentifiers
from substrafl.remote.register.generate_wheel import local_lib_wheels
from substrafl.remote.register.generate_wheel import pypi_lib_wheels
from substrafl.remote.remote_struct import RemoteStruct
from substrafl.remote.substratools_methods import RemoteMethod

logger = logging.getLogger(__name__)

# Substra tools version for which the image naming scheme changed
MINIMAL_DOCKER_SUBSTRATOOLS_VERSION = "0.16.0"

TMP_SUBSTRAFL_PREFIX = "tmp_substrafl"

_DEFAULT_SUBSTRATOOLS_IMAGE = "ghcr.io/substra/substra-tools:\
{substratools_version}-nvidiacuda11.8.0-base-ubuntu22.04-python{python_version}"

SUBSTRAFL_FOLDER = "substrafl_internal"

DOCKERFILE_TEMPLATE = """
FROM {docker_image}

COPY . .

# install dependencies
RUN python{python_version} -m pip install -U pip

# Install substrafl, substra (and substratools if editable mode)
{cl_deps}

# PyPi dependencies
{pypi_dependencies}

# Install local dependencies
{local_dependencies}

ENTRYPOINT ["python{python_version}", "algo.py", "--function-name", "{method_name}"]
"""

ALGO = """
import json
import cloudpickle

import substratools as tools

from substrafl.remote.remote_struct import RemoteStruct

from pathlib import Path

if __name__ == "__main__":
    # Load the wrapped user code
    remote_struct = RemoteStruct.load(src=Path(__file__).parent / '{substrafl_folder}')

    # Create a Substra algo from the wrapped user code
    remote_instance = remote_struct.get_remote_instance()

    # Register the functions to substra-tools
    remote_instance.register_substratools_function()

    # Execute the algo using substra-tools
    tools.execute()
"""


def _copy_local_packages(
    path: Path, local_dependencies: typing.List[Path], python_major_minor: str, operation_dir: Path
):
    """Copy the local libraries given by the user and generate the installation command."""
    dependencies_buffer = list()
    path.mkdir(exist_ok=True)
    for dependency_path in local_dependencies:
        dest_path = path / dependency_path.name
        if dependency_path.is_dir():
            shutil.copytree(
                dependency_path,
                dest_path,
                ignore=shutil.ignore_patterns(
                    "local-worker",
                    TMP_SUBSTRAFL_PREFIX + "*",
                ),
            )
        elif dependency_path.is_file():
            shutil.copy(dependency_path, dest_path)
        else:
            raise ValueError(f"Does not exist {dependency_path}")

        dependencies_buffer.append(f"{dest_path.relative_to(operation_dir)}")

    local_dependencies_cmd = (
        f"RUN python{python_major_minor} -m pip install --no-cache-dir " + " ".join(dependencies_buffer)
        if len(local_dependencies) > 0
        else ""
    )
    return local_dependencies_cmd


def _copy_local_code(path: Path, algo_file_path: Path, operation_dir: Path):
    """Copy the local code given by the user to the operation directory."""
    relative_path = path.relative_to(algo_file_path)
    (operation_dir / relative_path.parent).mkdir(exist_ok=True)
    if path.is_dir():
        shutil.copytree(path, operation_dir / relative_path)
    elif path.is_file():
        shutil.copy(path, operation_dir / relative_path)
    else:
        raise ValueError(f"Does not exist {path}")


def _create_archive(archive_path: Path, src_path: Path):
    """Create a tar archive from a folder"""
    with tarfile.open(archive_path, "w:gz") as tar:
        for filepath in src_path.glob("*"):
            if not filepath.name.endswith(".tar.gz"):
                tar.add(filepath, arcname=filepath.name, recursive=True)


def _get_base_docker_image(python_major_minor: str, editable_mode: bool):
    """Get the base Docker image for the Dockerfile"""

    substratools_image_version = substratools.__version__
    if util.strtobool(os.environ.get("USE_LATEST_SUBSTRATOOLS", "False")):
        substratools_image_version = "latest"
    elif version.parse(substratools_image_version) < version.parse(MINIMAL_DOCKER_SUBSTRATOOLS_VERSION):
        if not editable_mode:
            warnings.warn(
                f"Your environment uses substra-tools={substratools_image_version}. Version \
                {MINIMAL_DOCKER_SUBSTRATOOLS_VERSION} will be used on Docker.",
                exceptions.SubstraToolsDeprecationWarning,
            )
        substratools_image_version = MINIMAL_DOCKER_SUBSTRATOOLS_VERSION
    substratools_image = _DEFAULT_SUBSTRATOOLS_IMAGE.format(
        substratools_version=substratools_image_version,
        python_version=python_major_minor,
    )

    return substratools_image


def _create_substra_algo_files(
    remote_struct: RemoteStruct,
    install_libraries: bool,
    dependencies: Dependency,
    operation_dir: Path,
) -> typing.Tuple[Path, Path]:
    """Creates the necessary files from the remote struct to register the associated algorithm to substra, zip them into
        an archive (.tar.gz).

        Necessary files:

            - the RemoteStruct (wrapped code) dump file
            - the wheel of the current version of Substrafl and Substra
            - the Dockerfile
            - the description.md
            - the algo.py entrypoint

    Args:
        remote_struct (RemoteStruct): A representation of a substra algorithm.
        install_libraries (bool): whether we need to build the wheels and copy the files to install the libraries
        dependencies (Dependency): Algorithm dependencies.
        operation_dir (pathlib.Path): path to the operation directory

        Returns:
            Tuple[Path, Path]: The archive path and the description file path.
    """
    substrafl_internal = operation_dir / SUBSTRAFL_FOLDER
    substrafl_internal.mkdir()

    remote_struct.save(dest=substrafl_internal)

    # get Python version
    # Required to select the correct version of python inside the docker Image
    # Cloudpickle will crash if we don't deserialize with the same major.minor
    python_major_minor = ".".join(python_version().split(".")[:2])

    # Build Substrafl, Substra and Substratools wheel if needed
    install_cmd = ""

    if install_libraries:
        # Install either from pypi wheel or repo in editable mode
        if dependencies.editable_mode:
            install_cmd = local_lib_wheels(
                lib_modules=[
                    substrafl,
                    substra,
                    substratools,
                ],  # We reinstall substratools in editable mode to overwrite the installed version
                operation_dir=operation_dir,
                python_major_minor=python_major_minor,
                dest_dir=f"{SUBSTRAFL_FOLDER}/dist",
            )
        else:
            install_cmd = pypi_lib_wheels(
                lib_modules=[substrafl, substra],
                operation_dir=operation_dir,
                python_major_minor=python_major_minor,
                dest_dir=f"{SUBSTRAFL_FOLDER}/dist",
            )

    # Pypi dependencies docker command if specified by the user
    pypi_dependencies_cmd = (
        f"RUN python{python_major_minor} -m pip install --no-cache-dir {' '.join(dependencies.pypi_dependencies)}"
        if dependencies is not None and len(dependencies.pypi_dependencies) > 0
        else ""
    )

    # The files to copy to the container must be in the same folder as the Dockerfile
    local_dependencies_cmd = ""
    if dependencies is not None:
        algo_file_path = remote_struct.get_cls_file_path()
        for path in dependencies.local_code:
            _copy_local_code(path=path, algo_file_path=algo_file_path, operation_dir=operation_dir)

        if install_libraries:
            local_dep_dir = substrafl_internal / "local_dependencies"
            local_dependencies_cmd = _copy_local_packages(
                path=local_dep_dir,
                local_dependencies=dependencies.local_dependencies,
                python_major_minor=python_major_minor,
                operation_dir=operation_dir,
            )

    # Write template to algo.py
    algo_path = operation_dir / "algo.py"
    algo_path.write_text(
        ALGO.format(
            substrafl_folder=SUBSTRAFL_FOLDER,
        )
    )

    # Write description
    description_path = substrafl_internal / "description.md"
    description_path.write_text("# Substrafl Operation")

    # Get the base Docker image
    substratools_image = _get_base_docker_image(
        python_major_minor=python_major_minor, editable_mode=dependencies.editable_mode
    )

    # Write dockerfile based on template
    dockerfile_path = operation_dir / "Dockerfile"
    dockerfile_path.write_text(
        DOCKERFILE_TEMPLATE.format(
            docker_image=substratools_image,
            python_version=python_major_minor,
            cl_deps=install_cmd,
            pypi_dependencies=pypi_dependencies_cmd,
            local_dependencies=local_dependencies_cmd,
            method_name=remote_struct._method_name,
        )
    )

    # Create necessary archive to register the operation on substra
    archive_path = operation_dir / "algo.tar.gz"
    _create_archive(archive_path=archive_path, src_path=operation_dir)

    return archive_path, description_path


def register_algo(
    client: substra.Client,
    remote_struct: RemoteStruct,
    permissions: substra.sdk.schemas.Permissions,
    inputs: typing.List[substra.sdk.schemas.AlgoInputSpec],
    outputs: typing.List[substra.sdk.schemas.AlgoOutputSpec],
    dependencies: Dependency,
) -> str:
    """Automatically creates the needed files to register the composite algorithm associated to the remote_struct.

    Args:
        client (substra.Client): The substra client.
        remote_struct (RemoteStruct): The substra submittable algorithm representation.
        permissions (substra.sdk.schemas.Permissions): Permissions for the algorithm.
        inputs (typing.List[substra.sdk.schemas.AlgoInputSpec]): List of algo inputs to be used.
        outputs (typing.List[substra.sdk.schemas.AlgoOutputSpec]): List of algo outputs to be used.
        dependencies (Dependency): Algorithm dependencies.

    Returns:
        str: Substra algorithm key.
    """
    with tempfile.TemporaryDirectory(dir=str(Path.cwd().resolve()), prefix=TMP_SUBSTRAFL_PREFIX) as operation_dir:
        archive_path, description_path = _create_substra_algo_files(
            remote_struct,
            dependencies=dependencies,
            install_libraries=client.backend_mode != substra.BackendType.LOCAL_SUBPROCESS,
            operation_dir=Path(operation_dir),
        )
        key = client.add_algo(
            substra.sdk.schemas.AlgoSpec(
                name=remote_struct.algo_name,
                description=description_path,
                file=archive_path,
                inputs=inputs,
                outputs=outputs,
                permissions=permissions,
                metadata=dict(),
            )
        )
        return key


def _check_metric_function(metric_function: typing.Callable):
    """Function to check the type and the signature of a given metric function.

    Args:
        metric_function (typing.Callable): function to check.

    Raises:
        exceptions.MetricFunctionTypeError: metric_function must be of type "function"
        exceptions.MetricFunctionSignatureError: metric_function must ONLY contains
            datasamples and predictions_path as parameters
    """

    if not inspect.isfunction(metric_function):
        raise exceptions.MetricFunctionTypeError("The metric_function() must be of type function.")

    signature = inspect.signature(metric_function)
    parameters = signature.parameters

    if "datasamples" not in parameters:
        raise exceptions.MetricFunctionSignatureError(
            "The metric_function() function must contain datasamples as parameter."
        )
    elif "predictions_path" not in parameters:
        raise exceptions.MetricFunctionSignatureError(
            "The metric_function() function must contain predictions_path as parameter."
        )
    elif len(parameters) != 2:
        raise exceptions.MetricFunctionSignatureError(
            """The metric_function() function must ONLY contains datasamples and predictions_path as
            parameters."""
        )


def add_metric(
    client: substra.Client,
    permissions: substra.sdk.schemas.Permissions,
    dependencies: Dependency,
    metric_function: typing.Callable,
    metric_name: typing.Optional[str] = None,
) -> str:
    """Adds a metric to the Substra platform using the given metric function as the
    algorithm to execute.
    The metric function must be of type function, and its signature must ONLY contains
    `datasamples` and `predictions_path` as parameters. An error is raised otherwise.

    Args:
        client (substra.Client): The substra client.
        permissions (substra.sdk.schemas.Permissions): Permissions for the metric function.
        dependencies (Dependency): Metric function dependencies.
        metric_function (typing.Callable): function to compute the score from the datasamples and the predictions.
            This function is registered in substra as a metric.
        metric_name (str, Optional): Optional name chosen by the user to identify the metric. If None,
            the metric name is set to the 'metric_{metric_function.__name__}'.

    Returns:
        str: The metric key of the metric created from the metric function.
    """

    _check_metric_function(metric_function=metric_function)

    class Metric:
        def score(self, datasamples, predictions_path, _skip=True):
            # The _skip argument is needed to match the default signature of methods executed
            # on substratools_methods.py.
            return metric_function(datasamples=datasamples, predictions_path=predictions_path)

    inputs_metrics = [
        substra.sdk.schemas.AlgoInputSpec(
            identifier=InputIdentifiers.datasamples,
            kind=substra.sdk.schemas.AssetKind.data_sample,
            optional=False,
            multiple=True,
        ),
        substra.sdk.schemas.AlgoInputSpec(
            identifier=InputIdentifiers.opener,
            kind=substra.sdk.schemas.AssetKind.data_manager,
            optional=False,
            multiple=False,
        ),
        substra.sdk.schemas.AlgoInputSpec(
            identifier=InputIdentifiers.predictions,
            kind=substra.sdk.schemas.AssetKind.model,
            optional=False,
            multiple=False,
        ),
    ]

    outputs_metrics = [
        substra.sdk.schemas.AlgoOutputSpec(
            identifier=OutputIdentifiers.performance,
            kind=substra.sdk.schemas.AssetKind.performance,
            multiple=False,
        )
    ]

    remote_struct = RemoteStruct(
        cls=Metric,
        cls_args=[],
        cls_kwargs={},
        remote_cls=RemoteMethod,
        method_name="score",
        method_parameters={},
        algo_name=metric_name or "metric_" + metric_function.__name__,
    )

    key = register_algo(
        client=client,
        remote_struct=remote_struct,
        permissions=permissions,
        inputs=inputs_metrics,
        outputs=outputs_metrics,
        dependencies=dependencies,
    )

    return key



from typing import List
from typing import Optional

import numpy as np

from substrafl.algorithms.algo import Algo
from substrafl.nodes.aggregation_node import AggregationNode
from substrafl.nodes.references.local_state import LocalStateRef
from substrafl.nodes.references.shared_state import SharedStateRef
from substrafl.nodes.test_data_node import TestDataNode
from substrafl.nodes.train_data_node import TrainDataNode
from substrafl.remote import remote
from substrafl.schemas import ScaffoldAveragedStates
from substrafl.schemas import ScaffoldSharedState
from substrafl.schemas import StrategyName
from substrafl.strategies.strategy import Strategy


class Scaffold(Strategy):
    """Scaffold strategy.
    Paper: https://arxiv.org/pdf/1910.06378.pdf
    Scaffold is Federated Averaging with control variates.
    By adding auxiliary variables in the clients and server the authors of
    the related paper prove better bounds on the convergence assuming certain
    hypothesis, in particular with non-iid data.

    A round consists in performing a predefined number of forward/backward
    passes on each client, aggregating updates by computing their means and
    distributing the consensus update to all clients. In Scaffold, strategy is
    performed in a centralized way, where a single server or
    ``AggregationNode`` communicates with a number of clients ``TrainDataNode``
    and ``TestDataNode``.

    Args:
        aggregation_lr (float, Optional): Global aggregation rate applied on the averaged weight updates
            (`eta_g` in the paper). Defaults to 1. Must be >=0.
    """

    def __init__(self, aggregation_lr: float = 1):
        super(Scaffold, self).__init__()

        if aggregation_lr < 0:
            raise ValueError("aggregation_lr must be >=0")
        self._aggregation_lr = aggregation_lr
        # current local and share states references of the client used for training
        self._local_states: Optional[List[LocalStateRef]] = None
        self._shared_states: Optional[List[SharedStateRef]] = None

    @property
    def name(self) -> StrategyName:
        """The name of the strategy

        Returns:
            StrategyName: Name of the strategy
        """
        return StrategyName.SCAFFOLD

    def perform_round(
        self,
        algo: Algo,
        train_data_nodes: List[TrainDataNode],
        aggregation_node: AggregationNode,
        round_idx: int,
        clean_models: bool,
        additional_orgs_permissions: Optional[set] = None,
    ):
        """One round of the Scaffold strategy consists in:
            - if ``round_idx==0``: initialize the strategy by performing a local update
                (train on n mini-batches) of the models on each train data nodes
            - aggregate the model shared_states
            - set the model weights to the aggregated weights on each train data nodes
            - perform a local update (train on n mini-batches) of the models on each train data nodes

        Args:
            algo (Algo): User defined algorithm: describes the model train and predict methods
            train_data_nodes (typing.List[TrainDataNode]): List of the organizations on which to perform
            local updates aggregation_node (AggregationNode): Node without data, used to perform
                operations on the shared states of the models
            round_idx (int): Round number, it starts at 0.
            clean_models (bool): Clean the intermediary models of this round on the Substra platform.
                Set it to False if you want to download or re-use intermediary models. This causes the disk
                space to fill quickly so should be set to True unless needed.
            additional_orgs_permissions (typing.Optional[set]): Additional permissions to give to the model outputs
                after training, in order to test the model on an other organization.
        """
        if aggregation_node is None:
            raise ValueError("In Scaffold strategy aggregation node cannot be None")

        if round_idx == 0:
            # Initialization of the strategy by performing a local update on each train data organization
            assert self._local_states is None
            assert self._shared_states is None
            self._perform_local_updates(
                algo=algo,
                train_data_nodes=train_data_nodes,
                current_aggregation=None,
                round_idx=round_idx,
                aggregation_id=aggregation_node.organization_id,
                additional_orgs_permissions=additional_orgs_permissions or set(),
                clean_models=clean_models,
            )

        else:
            current_aggregation = aggregation_node.update_states(
                self.avg_shared_states(shared_states=self._shared_states, _algo_name="Aggregating"),  # type: ignore
                round_idx=round_idx,
                authorized_ids=set([train_data_node.organization_id for train_data_node in train_data_nodes]),
                clean_models=clean_models,
            )

            self._perform_local_updates(
                algo=algo,
                train_data_nodes=train_data_nodes,
                current_aggregation=current_aggregation,
                round_idx=round_idx,
                aggregation_id=aggregation_node.organization_id,
                additional_orgs_permissions=additional_orgs_permissions or set(),
                clean_models=clean_models,
            )

    def predict(
        self,
        algo: Algo,
        test_data_nodes: List[TestDataNode],
        train_data_nodes: List[TrainDataNode],
        round_idx: int,
    ):

        for test_data_node in test_data_nodes:
            matching_train_nodes = [
                train_data_node
                for train_data_node in train_data_nodes
                if train_data_node.organization_id == test_data_node.organization_id
            ]
            if len(matching_train_nodes) == 0:
                node_index = 0
            else:
                node_index = train_data_nodes.index(matching_train_nodes[0])

            assert self._local_states is not None, "Cannot predict if no training has been done beforehand."
            local_state = self._local_states[node_index]

            test_data_node.update_states(
                operation=algo.predict(
                    data_samples=test_data_node.test_data_sample_keys,
                    shared_state=None,
                    _algo_name=f"Testing with {algo.__class__.__name__}",
                ),
                traintask_id=local_state.key,
                round_idx=round_idx,
            )  # Init state for testtask

    def _check_shared_states(self, shared_states: List[ScaffoldSharedState]):
        """Check the Scaffold assumptions: server_control_variate, parameters_update and server_control_variate have the
        same length for all the shared states, and all server_control_variate are equal.

        Args:
            shared_states (List[ScaffoldSharedState]): Shared state returned by the train method of the algorithm for
                each client (e.g. algorithms.pytorch.scaffold.train)
        """
        assert shared_states, "shared_states should contain at least one element"

        for shared_state in shared_states:

            assert isinstance(
                shared_state, ScaffoldSharedState
            ), "shared_state should be an instance of ScaffoldSharedState"

            assert len(shared_state.control_variate_update) == len(
                shared_states[0].control_variate_update
            ), "the length of control_variate_update should be the same for each shared_state"
            assert len(shared_state.parameters_update) == len(
                shared_states[0].parameters_update
            ), "the length of parameters_update should be the same for each shared_state"
            assert len(shared_state.server_control_variate) == len(
                shared_states[0].server_control_variate
            ), "the length of server_control_variate should be the same for each shared_state"

            for c, ci in zip(shared_states[0].server_control_variate, shared_state.server_control_variate):
                np.testing.assert_array_equal(
                    c, ci, err_msg="all server_control_variate in the shared_states are not equal"
                )

        assert (
            len(shared_states[0].control_variate_update)
            == len(shared_states[0].server_control_variate)
            == len(shared_states[0].parameters_update)
        ), "the length of server_control_variate, parameters_update and server_control_variate should be the same"

    def _weight_arrays(
        self,
        client_weight: np.ndarray,
        states_to_aggregate: List[List[np.ndarray]],
        layer_idx: int,
    ) -> List[np.ndarray]:
        """apply the `client_weight` to the arrays of the layer `layer_idx` in `states_to_aggregate`
        Args:
            client_weight (numpy.ndarray): array of shape (num_clients,). Contains the weight of
                each client (n_samples / n_all_samples).
            states_to_aggregate (typing.List[typing.List[numpy.ndarray]]): List of the states on
                which the weights are applied.
            layer_idx (int): the selected layer

        Returns:
            typing.List[numpy.ndarray]: the weighted arrays of the layer `layer_idx`
        """
        weighted_arrays = []
        assert len(client_weight) == len(
            states_to_aggregate
        ), "n_samples_per_client and states_to_aggregate should have the same length"

        for client_idx in range(len(states_to_aggregate)):
            layer_values = states_to_aggregate[client_idx][layer_idx]
            weighted_array = client_weight[client_idx] * layer_values
            weighted_arrays.append(weighted_array)

        return weighted_arrays

    def _update_server_control_variate(
        self,
        server_control_variate: List[np.ndarray],
        control_variate_updates: List[List[np.ndarray]],
        client_weight: np.ndarray,
    ) -> List[np.ndarray]:
        """Updates the server control variate with the weighted average of the control variate updates,
        according to Scaffold paper's Algo steps 16.1 + 17.1: `c = c + sum([client_weight*control_variate_update])`

        Args:
            server_control_variate (typing.List[numpy.ndarray]): the server control variate
            control_variate_updates (typing.List[typing.List[numpy.ndarray]]): the control variate updates from
                the shared state of each client.
            client_weight (numpy.ndarray): array of shape (num_clients,). Contains the weight of each client
                (n_samples / n_all_samples).

        Returns:
            typing.List[numpy.ndarray]: the updated server_control_variate
        """
        updated_server_control_variate = []
        # aggregate the arrays at pos i of each client state
        for layer_idx in range(len(control_variate_updates[0])):
            weighted_arrays = self._weight_arrays(
                client_weight=client_weight,
                states_to_aggregate=control_variate_updates,
                layer_idx=layer_idx,
            )
            # get array of position i for each client state and multiply by the client weight

            weighted_arrays.append(server_control_variate[layer_idx])
            updated_server_control_variate.append(np.sum(weighted_arrays, axis=0))

        return updated_server_control_variate

    def _avg_weight_update(
        self,
        weight_updates: List[List[np.ndarray]],
        client_weight: np.ndarray,
    ) -> List[np.ndarray]:
        """Computes the weighted average of the weight updates and applies the aggregation learning rate,
        according to Scaffold paper's Algo steps 16.2 + 17.2:
        `delta_x = global_lr * sum([client_weight*parameters_update])`

        Args:
            weight_updates (typing.List[typing.List[numpy.ndarray]]): the weight updates of the clients
            client_weight (numpy.ndarray): array of shape (num_clients,). Contains the weight of each client
                (n_samples / n_all_samples).

        Returns:
            typing.List[numpy.ndarray]: the averaged weight updates
        """
        averaged_weight_update = []
        # aggregate the arrays at pos i of each client state
        for layer_idx in range(len(weight_updates[0])):
            weighted_arrays = self._weight_arrays(
                client_weight=client_weight,
                states_to_aggregate=weight_updates,
                layer_idx=layer_idx,
            )
            # we apply global_lr here so we don't have to pass it to the algo (step 17.2)
            averaged_weight_update.append(self._aggregation_lr * np.sum(weighted_arrays, axis=0))

        return averaged_weight_update

    @remote
    def avg_shared_states(self, shared_states: List[ScaffoldSharedState]) -> ScaffoldAveragedStates:
        """Performs the aggregation of the shared states returned by the train
        methods of the user-defined algorithm, according to the server operations of the Scaffold Algo.

        1. Computes the weighted average of the weight updates and applies the aggregation learning rate
        2. Updates the server control variate with the weighted average of the control variate updates

        The average is weighted by the proportion of the number of samples.

        Args:
            shared_states (typing.List[ScaffoldSharedState]): Shared state returned by the train method of
                the algorithm for each client (e.g. algorithms.pytorch.scaffold.train)

        Returns:
            ScaffoldAveragedStates: averaged weight updates and updated server control variate
        """
        # TODO: Do separate function for pytorch to avoid converting in np

        self._check_shared_states(shared_states=shared_states)

        # remove "n_samples" from shared_states and store it in all_samples
        n_samples_per_client = np.array([state.n_samples for state in shared_states])
        client_weight = n_samples_per_client / np.sum(n_samples_per_client)

        # all values should be the same: take the first one
        server_control_variate = shared_states[0].server_control_variate

        averaged_states = ScaffoldAveragedStates(
            server_control_variate=self._update_server_control_variate(
                server_control_variate=server_control_variate,
                control_variate_updates=[shared_state.control_variate_update for shared_state in shared_states],
                client_weight=client_weight,
            ),
            avg_parameters_update=self._avg_weight_update(
                weight_updates=[shared_state.parameters_update for shared_state in shared_states],
                client_weight=client_weight,
            ),
        )

        return averaged_states

    def _perform_local_updates(
        self,
        algo: Algo,
        train_data_nodes: List[TrainDataNode],
        current_aggregation: Optional[SharedStateRef],
        round_idx: int,
        aggregation_id: str,
        additional_orgs_permissions: set,
        clean_models: bool,
    ):
        """Perform a local update (train on n mini-batches) of the models
        on each train data nodes.

        Args:
            algo (Algo): User defined algorithm: describes the model train and predict methods
            train_data_nodes (typing.List[TrainDataNode]): List of the organizations on which to perform
            local updates current_aggregation (SharedStateRef, Optional): Reference of an aggregation operation to be
            passed as input to each local training
            round_idx (int): Round number, it starts at 1.
            aggregation_id (str): Id of the aggregation node the shared state is given to.
            additional_orgs_permissions (set): Additional permissions to give to the model outputs
                after training, in order to test the model on an other organization.
            clean_models (bool): Clean the intermediary models of this round on the Substra platform.
                Set it to False if you want to download or re-use intermediary models. This causes the disk
                space to fill quickly so should be set to True unless needed.
        """

        next_local_states = []
        next_shared_states = []

        for i, node in enumerate(train_data_nodes):
            # define composite tasks (do not submit yet)
            # for each composite task give description of Algo instead of a key for an algo
            next_local_state, next_shared_state = node.update_states(
                algo.train(  # type: ignore
                    node.data_sample_keys,
                    shared_state=current_aggregation,
                    _algo_name=f"Training with {algo.__class__.__name__}",
                ),
                local_state=self._local_states[i] if self._local_states is not None else None,
                round_idx=round_idx,
                authorized_ids=set([node.organization_id]) | additional_orgs_permissions,
                aggregation_id=aggregation_id,
                clean_models=clean_models,
            )
            # keep the states in a list: one/organization
            next_local_states.append(next_local_state)
            next_shared_states.append(next_shared_state)

        self._local_states = next_local_states
        self._shared_states = next_shared_states



from typing import Generator
from typing import List

import torch


def is_batchnorm_layer(layer: torch.nn.Module) -> bool:
    """Checks if the provided layer is a Batch Norm layer (either 1D, 2D or 3D).

    Args:
        layer (torch.nn.Module): Pytorch module.

    Returns:
        bool: Whether the given module is a batch norm one.
    """
    list_bn_layers = [
        torch.nn.BatchNorm1d,
        torch.nn.BatchNorm2d,
        torch.nn.BatchNorm3d,
        torch.nn.LazyBatchNorm1d,
        torch.nn.LazyBatchNorm2d,
        torch.nn.LazyBatchNorm3d,
    ]
    for bn_layer_class in list_bn_layers:
        if isinstance(layer, bn_layer_class):
            return True

    return False


def batch_norm_param(
    model: torch.nn.Module,
) -> Generator[torch.nn.parameter.Parameter, None, None]:
    """Generator of the internal parameters of the batch norm layers
    of the model. This yields references hence all modification done to the yielded object will
    be applied to the input model.

    The internal parameters of a batch norm layer include the running mean and
    the running variance.

    Args:
        model (torch.nn.Module): A torch model.

    Yields:
        float: The running mean and variance of all batch norm layers parameters from the given model.
    """
    for _, module in model.named_modules():
        if is_batchnorm_layer(module):
            yield module.running_mean
            yield module.running_var


def model_parameters(model: torch.nn.Module, with_batch_norm_parameters: bool) -> torch.nn.parameter.Parameter:
    """A generator of the given model parameters. The returned generator yields references hence all modification done
    to the yielded object will be applied to the input model. If with_batch_norm_parameters is set to True, the running
    mean and the running variance of each batch norm layer will be added after the "classic" parameters.

    Args:
        model (torch.nn.Module): A torch model.
        with_batch_norm_parameters (bool): If set to True, the running mean
            and the running variance of each batch norm layer will be added
            after the "classic" parameters.

    Returns:
        typing.Generator[torch.nn.parameter.Parameter, typing.Any, typing.Any]: A python generator of torch parameters.
    """

    def my_iterator():
        for p in model.parameters():
            yield p

        if with_batch_norm_parameters:
            for p in batch_norm_param(model):
                yield p

    return my_iterator


def get_parameters(
    model: torch.nn.Module,
    with_batch_norm_parameters: bool,
) -> List[torch.nn.parameter.Parameter]:
    """Model parameters from the provided torch model. This function returns a copy not a reference. If
    with_batch_norm_parameters is set to True, the running mean and the running variance of the batch norm
    layers will be added after the "classic" parameters of the model.

    Args:
        model (torch.nn.Module): A torch model.
        with_batch_norm_parameters (bool): If set to True, the running mean
            and the running variance of each batch norm layer will be added
            after the "classic" parameters.

    Returns:
        typing.List[torch.nn.parameter.Parameter]: The list of torch parameters of the provided model.
    """
    with torch.inference_mode():
        iter_params = model_parameters(model, with_batch_norm_parameters=with_batch_norm_parameters)
        parameters = [p.clone() for p in iter_params()]

    return parameters


def increment_parameters(
    model: torch.nn.Module,
    updates: List[torch.nn.parameter.Parameter],
    with_batch_norm_parameters: bool,
    updates_multiplier: float = 1.0,
):
    """Add the given update to the model parameters. If with_batch_norm_parameters is set to True, the operation
    will include the running mean and the running variance of the batch norm layers (in this case, they must be
    included in the given update). This function modifies the given model internally and therefore returns nothing.

    Args:
        model (torch.nn.Module): The torch model to modify.
        updates (typing.List[torch.nn.parameter.Parameter]): A list of torch parameters
            to add to the model, as ordered by the standard iterators. The trainable parameters should come first
            followed by the batch norm parameters if `with_batch_norm_parameters` is set to `True`.
            If the type is np.ndarray, it is converted in `torch.Tensor`.
        with_batch_norm_parameters (bool): If set to True, the running mean and the running variance of
            each batch norm layer will be included, after the trainable layers, in the model parameters to modify.
        updates_multiplier (float, Optional): The coefficient which multiplies the updates before being added to the
            model. Defaults to 1.0.
    """
    with torch.inference_mode():
        # INFO: this is the faster way I found of checking that both model.parameters() and shared states has the
        # same length as model.parameters() is a generator.
        iter_params = model_parameters(model=model, with_batch_norm_parameters=with_batch_norm_parameters)
        n_parameters = len(list(iter_params()))
        assert n_parameters == len(updates), "Length of model parameters and updates are unequal."

        for weights, update in zip(iter_params(), updates):
            assert update.data.shape == weights.data.shape, (
                f"The shape of the model weights ({weights.data.shape}) and of the update ({update.data.shape}) "
                "passed in the updates argument are unequal."
            )
            weights.data += updates_multiplier * update.data


def subtract_parameters(
    parameters: List[torch.nn.parameter.Parameter],
    parameters_to_subtract: List[torch.nn.parameter.Parameter],
) -> List[torch.nn.parameter.Parameter]:
    """
    subtract the given list of torch parameters i.e. : parameters - parameters_to_subtract.
    Those elements can be extracted from a model thanks to the :func:`~get_parameters` function.

    Args:
        parameters (typing.List[torch.nn.parameter.Parameter]): A list of torch parameters.
        parameters_to_subtract (typing.List[torch.nn.parameter.Parameter]): A list of torch parameters.

    Returns:
        typing.List[torch.nn.parameter.Parameter]: The subtraction of the given parameters.
    """
    return weighted_sum_parameters(
        parameters_list=[parameters, parameters_to_subtract],
        coefficient_list=[1, -1],
    )


def add_parameters(
    parameters: List[torch.nn.parameter.Parameter],
    parameters_to_add: List[torch.nn.parameter.Parameter],
) -> List[torch.nn.parameter.Parameter]:
    """
    add the given list of torch parameters i.e. : parameters - parameters_to_add.
    Those elements can be extracted from a model thanks to the :func:`~get_parameters` function.

    Args:
        parameters (typing.List[torch.nn.parameter.Parameter]): A list of torch parameters.
        parameters_to_add (typing.List[torch.nn.parameter.Parameter]): A list of torch parameters.

    Returns:
        typing.List[torch.nn.parameter.Parameter]: The addition of the given parameters.
    """
    return weighted_sum_parameters(
        parameters_list=[parameters, parameters_to_add],
        coefficient_list=[1, 1],
    )


def weighted_sum_parameters(
    parameters_list: List[List[torch.Tensor]],
    coefficient_list: List[float],
) -> List[torch.Tensor]:
    """
    Do a weighted sum of the given lists of torch parameters.
    Those elements can be extracted from a model thanks to the :func:`~get_parameters` function.

    Args:
        parameters_list (typing.List[List[torch.Tensor]]): A list of List of torch parameters.
        coefficient_list (typing.List[float]): A list of coefficients which will be applied to each list of parameters.
    Returns:
        typing.List[torch.nn.parameter.Parameter]: The weighted sum of the given list of torch parameters.
    """

    weighted_sum = []

    assert all(
        len(parameters_list[0]) == len(parameters) for parameters in parameters_list
    ), "The number of parameters in each List is not the same"

    assert len(parameters_list) == len(coefficient_list), "There must be a coefficient for each List of parameters"

    for parameters_to_sum in zip(*parameters_list):
        assert all(
            parameters_to_sum[0].data.shape == parameter.data.shape for parameter in parameters_to_sum
        ), "The shape of the parameters are unequal."
        with torch.inference_mode():
            weighted_sum.append(sum(param * coeff for param, coeff in zip(parameters_to_sum, coefficient_list)))

    return weighted_sum


def set_parameters(
    model: torch.nn.Module,
    parameters: List[torch.nn.parameter.Parameter],
    with_batch_norm_parameters: bool,
):
    """Sets the parameters of a pytorch model to the provided parameters. If with_batch_norm_parameters is set to True,
    the operation will include the running mean and the running variance of the batch norm layers (in this case, they
    must be included in the given parameters). This function modifies the given model internally and therefore returns
    nothing.

    Args:
        model (torch.nn.Module): The torch model to modify.
        parameters (typing.List[torch.nn.parameter.Parameter]): Model parameters, as ordered by the standard iterators.
        The trainable parameters should come first followed by the batch norm parameters if `with_batch_norm_parameters`
            is set to `True`.
        with_batch_norm_parameters (bool): Whether to the batch norm layers' internal parameters are provided and
            need to be included in the operation.
    """
    with torch.inference_mode():
        iter_params = model_parameters(model, with_batch_norm_parameters=with_batch_norm_parameters)
        n_parameters = len(list(iter_params()))
        assert n_parameters == len(parameters), "Length of model parameters and provided parameters are unequal."
        for (p, w) in zip(iter_params(), parameters):
            p.data = w.data


def zeros_like_parameters(
    model: torch.nn.Module,
    with_batch_norm_parameters: bool,
    device: torch.device,
) -> List[torch.Tensor]:
    """Copy the model parameters from the provided torch model and sets values to zero.
    If with_batch_norm_parameters is set to True, the running mean and the running variance of the batch norm
    layers will be added after the "classic" parameters of the model.

    Args:
        model (torch.nn.Module): A torch model.
        with_batch_norm_parameters (bool): If set to True, the running mean
            and the running variance of each batch norm layer will be added
            after the "classic" parameters.
        device (torch.device): torch device on which to save the parameters

    Returns:
        typing.List[torch.nn.parameter.Parameter]: The list of torch parameters of the provided model
        with values set to zero.
    """
    with torch.inference_mode():
        iter_params = model_parameters(model, with_batch_norm_parameters=with_batch_norm_parameters)
        parameters = [torch.zeros_like(p).to(device) for p in iter_params()]

    return parameters



from abc import ABC
from abc import abstractmethod
from pathlib import Path
from typing import Any


class Serializer(ABC):
    @staticmethod
    @abstractmethod
    def save(state: Any, path: Path):
        """Save the state to the path

        Args:
            state (typing.Any): state to save
            path (pathlib.Path): path where to save it
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def load(path: Path):
        raise NotImplementedError



from substrafl.strategies.fed_avg import FedAvg
from substrafl.strategies.newton_raphson import NewtonRaphson
from substrafl.strategies.scaffold import Scaffold
from substrafl.strategies.single_organization import SingleOrganization
from substrafl.strategies.strategy import Strategy

__all__ = ["Strategy", "FedAvg", "SingleOrganization", "Scaffold", "NewtonRaphson"]



import uuid
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

import substra
from substra import schemas

from substrafl.dependency import Dependency
from substrafl.nodes.node import InputIdentifiers
from substrafl.nodes.node import Node
from substrafl.nodes.node import OperationKey
from substrafl.nodes.node import OutputIdentifiers
from substrafl.nodes.references.local_state import LocalStateRef
from substrafl.nodes.references.shared_state import SharedStateRef
from substrafl.remote.operations import DataOperation
from substrafl.remote.register import register_algo
from substrafl.remote.remote_struct import RemoteStruct


class TrainDataNode(Node):
    """
    A predefined structure that allows you to register operations
    on your train node in a static way before submitting them to substra.

    Args:
        organization_id (str): The substra organization ID (shared with other organizations if permissions are needed)
        data_manager_key (str): Substra data_manager_key opening data samples used by the strategy
        data_sample_keys (typing.List[str]): Substra data_sample_keys used for the training on this node
    """

    def __init__(
        self,
        organization_id: str,
        data_manager_key: str,
        data_sample_keys: List[str],
    ):
        self.data_manager_key = data_manager_key
        self.data_sample_keys = data_sample_keys

        super(TrainDataNode, self).__init__(organization_id)

    def update_states(
        self,
        operation: DataOperation,
        round_idx: int,
        authorized_ids: Set[str],
        aggregation_id: Optional[str] = None,
        clean_models: bool = False,
        local_state: Optional[LocalStateRef] = None,
    ) -> Tuple[LocalStateRef, SharedStateRef]:
        """Adding a composite train task to the list of operations to
        be executed by the node during the compute plan. This is done in a static
        way, nothing is submitted to substra.
        This is why the algo key is a RemoteStruct (substrafl local reference of the algorithm)
        and not a substra algo_key as nothing has been submitted yet.

        Args:
            operation (DataOperation): Automatically generated structure returned by
                the :py:func:`~substrafl.remote.decorators.remote_data` decorator. This allows to register an
                operation and execute it later on.
            round_idx (int): Round number, it starts at 1. In case of a centralized strategy,
                it is preceded by an initialization round tagged: 0.
            authorized_ids (typing.Set[str]): Authorized org to access the output model.
            aggregation_id (str): Aggregation node id to authorize access to the shared model. Default to None.
            clean_models (bool): Whether outputs of this operation are transient (deleted when they are not used
                anymore) or not. Defaults to False.
            local_state (typing.Optional[LocalStateRef]): The parent task LocalStateRef. Defaults to None.

        Raises:
            TypeError: operation must be a DataOperation, make sure to decorate the train and predict methods of
                your algorithm with @remote

        Returns:
            typing.Tuple[LocalStateRef, SharedStateRef]: Identifications for the results of this operation.
        """
        if not isinstance(operation, DataOperation):
            raise TypeError(
                "operation must be a DataOperation",
                f"Given: {type(operation)}",
                "Have you decorated your method with @remote_data?",
            )

        op_id = str(uuid.uuid4())
        data_inputs = [schemas.InputRef(identifier=InputIdentifiers.opener, asset_key=self.data_manager_key)] + [
            schemas.InputRef(identifier=InputIdentifiers.datasamples, asset_key=data_sample)
            for data_sample in self.data_sample_keys
        ]

        local_inputs = (
            [
                schemas.InputRef(
                    identifier=InputIdentifiers.local,
                    parent_task_key=local_state.key,
                    parent_task_output_identifier=OutputIdentifiers.local,
                )
            ]
            if local_state is not None
            else []
        )

        if operation.shared_state is not None:
            shared_inputs = [
                schemas.InputRef(
                    identifier=InputIdentifiers.shared,
                    parent_task_key=operation.shared_state.key,
                    parent_task_output_identifier=OutputIdentifiers.model,
                )
            ]

        elif local_state is not None:
            shared_inputs = [
                schemas.InputRef(
                    identifier=InputIdentifiers.shared,
                    parent_task_key=local_state.key,
                    parent_task_output_identifier=OutputIdentifiers.shared,
                )
            ]

        else:
            shared_inputs = []

        composite_traintask = schemas.ComputePlanTaskSpec(
            algo_key=str(uuid.uuid4()),  # bogus algo key
            task_id=op_id,
            inputs=data_inputs + local_inputs + shared_inputs,
            outputs={
                OutputIdentifiers.shared: schemas.ComputeTaskOutputSpec(
                    permissions=schemas.Permissions(
                        public=False,
                        authorized_ids=list(authorized_ids | set([aggregation_id] if aggregation_id else [])),
                    ),
                    transient=clean_models,
                ),
                OutputIdentifiers.local: schemas.ComputeTaskOutputSpec(
                    permissions=schemas.Permissions(public=False, authorized_ids=list(authorized_ids)),
                    transient=clean_models,
                ),
            },
            metadata={
                "round_idx": round_idx,
            },
            tag="train",
            worker=self.organization_id,
        ).dict()

        composite_traintask.pop("algo_key")
        composite_traintask["remote_operation"] = operation.remote_struct

        self.tasks.append(composite_traintask)

        return LocalStateRef(op_id), SharedStateRef(op_id)

    def register_operations(
        self,
        client: substra.Client,
        permissions: schemas.Permissions,
        cache: Dict[RemoteStruct, OperationKey],
        dependencies: Dependency,
    ) -> Dict[RemoteStruct, OperationKey]:
        """Define the algorithms for each operation and submit the composite traintask to substra.

        Go through every operation in the computation graph, check what algorithm they use (identified by their
        RemoteStruct id), submit it to substra and save `RemoteStruct : algo_key` into the `cache` (where algo_key
        is the returned algo key by substra.)
        If two tasks depend on the same algorithm, the algorithm won't be added twice to substra as this method check
        if an algo has already been submitted to substra before adding it.

        Args:
            client (substra.Client): Substra client for the organization.
            permissions (substra.sdk.schemas.Permissions): schemas.Permissions for the algorithm.
            cache (typing.Dict[RemoteStruct, OperationKey]): Already registered algorithm identifications. The key of
                each element is the RemoteStruct id (generated by substrafl) and the value is the key generated by
                substra.
            dependencies (Dependency): Algorithm dependencies.

        Returns:
            typing.Dict[RemoteStruct, OperationKey]: updated cache
        """
        for task in self.tasks:
            if isinstance(task["remote_operation"], RemoteStruct):
                remote_struct: RemoteStruct = task["remote_operation"]

                if remote_struct not in cache:
                    algo_key = register_algo(
                        client=client,
                        remote_struct=remote_struct,
                        permissions=permissions,
                        inputs=[
                            schemas.AlgoInputSpec(
                                identifier=InputIdentifiers.datasamples,
                                kind=schemas.AssetKind.data_sample.value,
                                optional=False,
                                multiple=True,
                            ),
                            schemas.AlgoInputSpec(
                                identifier=InputIdentifiers.opener,
                                kind=schemas.AssetKind.data_manager.value,
                                optional=False,
                                multiple=False,
                            ),
                            schemas.AlgoInputSpec(
                                identifier=InputIdentifiers.local,
                                kind=schemas.AssetKind.model.value,
                                optional=True,
                                multiple=False,
                            ),
                            schemas.AlgoInputSpec(
                                identifier=InputIdentifiers.shared,
                                kind=schemas.AssetKind.model.value,
                                optional=True,
                                multiple=False,
                            ),
                        ],
                        outputs=[
                            schemas.AlgoOutputSpec(
                                identifier=OutputIdentifiers.local, kind=schemas.AssetKind.model.value, multiple=False
                            ),
                            schemas.AlgoOutputSpec(
                                identifier=OutputIdentifiers.shared, kind=schemas.AssetKind.model.value, multiple=False
                            ),
                        ],
                        dependencies=dependencies,
                    )
                    cache[remote_struct] = algo_key
                else:
                    algo_key = cache[remote_struct]

                task["algo_key"] = algo_key

        return cache

    def summary(self) -> dict:
        """Summary of the class to be exposed in the experiment summary file

        Returns:
            dict: a json-serializable dict with the attributes the user wants to store
        """
        summary = super().summary()
        summary.update(
            {
                "data_manager_key": self.data_manager_key,
                "data_sample_keys": self.data_sample_keys,
            }
        )
        return summary



from typing import List
from typing import Optional

import numpy as np

from substrafl.algorithms.algo import Algo
from substrafl.exceptions import EmptySharedStatesError
from substrafl.nodes.aggregation_node import AggregationNode
from substrafl.nodes.references.local_state import LocalStateRef
from substrafl.nodes.references.shared_state import SharedStateRef
from substrafl.nodes.test_data_node import TestDataNode
from substrafl.nodes.train_data_node import TrainDataNode
from substrafl.remote import remote
from substrafl.schemas import FedAvgAveragedState
from substrafl.schemas import FedAvgSharedState
from substrafl.schemas import StrategyName
from substrafl.strategies.strategy import Strategy


class FedAvg(Strategy):
    """Federated averaging strategy.

    Federated averaging is the simplest federating strategy.
    A round consists in performing a predefined number of forward/backward
    passes on each client, aggregating updates by computing their means and
    distributing the consensus update to all clients. In FedAvg, strategy is
    performed in a centralized way, where a single server or
    ``AggregationNode`` communicates with a number of clients ``TrainDataNode``
    and ``TestDataNode``.

    Formally, if :math:`w_t` denotes the parameters of the model at round
    :math:`t`, a single round consists in the following steps:

    .. math::

      \\Delta w_t^{k} = \\mathcal{O}^k_t(w_t| X_t^k, y_t^k, m)
      \\Delta w_t = \\sum_{k=1}^K \\frac{n_k}{n} \\Delta w_t^k
      w_{t + 1} = w_t + \\Delta w_t

    where :math:`\\mathcal{O}^k_t` is the local optimizer algorithm of client
    :math:`k` taking as argument the averaged weights as well as the
    :math:`t`-th batch of data for local worker :math:`k` and the number of
    local updates :math:`m` to perform, and where :math:`n_k` is the number of
    samples for worker :math:`k`, :math:`n = \\sum_{k=1}^K n_k` is the total
    number of samples.
    """

    def __init__(self):
        super(FedAvg, self).__init__()

        # current local and share states references of the client
        self._local_states: Optional[List[LocalStateRef]] = None
        self._shared_states: Optional[List[SharedStateRef]] = None

    @property
    def name(self) -> StrategyName:
        """The name of the strategy

        Returns:
            StrategyName: Name of the strategy
        """
        return StrategyName.FEDERATED_AVERAGING

    def perform_round(
        self,
        algo: Algo,
        train_data_nodes: List[TrainDataNode],
        aggregation_node: AggregationNode,
        round_idx: int,
        clean_models: bool,
        additional_orgs_permissions: Optional[set] = None,
    ):
        """One round of the Federated Averaging strategy consists in:
            - if ``round_idx==0``: initialize the strategy by performing a local update
                (train on n mini-batches) of the models on each train data node
            - aggregate the model shared_states
            - set the model weights to the aggregated weights on each train data nodes
            - perform a local update (train on n mini-batches) of the models on each train data nodes

        Args:
            algo (Algo): User defined algorithm: describes the model train and predict methods
            train_data_nodes (typing.List[TrainDataNode]): List of the nodes on which to perform
                local updates.
            aggregation_node (AggregationNode): Node without data, used to perform
                operations on the shared states of the models
            round_idx (int): Round number, it starts at 0.
            clean_models (bool): Clean the intermediary models of this round on the Substra platform.
                Set it to False if you want to download or re-use intermediary models. This causes the disk
                space to fill quickly so should be set to True unless needed.
            additional_orgs_permissions (typing.Optional[set]): Additional permissions to give to the model outputs
                after training, in order to test the model on an other organization.
        """
        if aggregation_node is None:
            raise ValueError("In FedAvg strategy aggregation node cannot be None")

        if round_idx == 0:
            # Initialization of the strategy by performing a local update on each train data organization
            assert self._local_states is None
            assert self._shared_states is None
            self._perform_local_updates(
                algo=algo,
                train_data_nodes=train_data_nodes,
                current_aggregation=None,
                round_idx=round_idx,
                aggregation_id=aggregation_node.organization_id,
                additional_orgs_permissions=additional_orgs_permissions or set(),
                clean_models=clean_models,
            )

        else:
            current_aggregation = aggregation_node.update_states(
                self.avg_shared_states(shared_states=self._shared_states, _algo_name="Aggregating"),  # type: ignore
                round_idx=round_idx,
                authorized_ids=set([train_data_node.organization_id for train_data_node in train_data_nodes]),
                clean_models=clean_models,
            )

            self._perform_local_updates(
                algo=algo,
                train_data_nodes=train_data_nodes,
                current_aggregation=current_aggregation,
                round_idx=round_idx,
                aggregation_id=aggregation_node.organization_id,
                additional_orgs_permissions=additional_orgs_permissions or set(),
                clean_models=clean_models,
            )

    def predict(
        self,
        algo: Algo,
        test_data_nodes: List[TestDataNode],
        train_data_nodes: List[TrainDataNode],
        round_idx: int,
    ):

        for test_data_node in test_data_nodes:
            matching_train_nodes = [
                train_data_node
                for train_data_node in train_data_nodes
                if train_data_node.organization_id == test_data_node.organization_id
            ]
            if len(matching_train_nodes) == 0:
                node_index = 0
            else:
                node_index = train_data_nodes.index(matching_train_nodes[0])

            assert self._local_states is not None, "Cannot predict if no training has been done beforehand."
            local_state = self._local_states[node_index]

            test_data_node.update_states(
                traintask_id=local_state.key,
                operation=algo.predict(
                    data_samples=test_data_node.test_data_sample_keys,
                    shared_state=None,
                    _algo_name=f"Testing with {algo.__class__.__name__}",
                ),
                round_idx=round_idx,
            )  # Init state for testtask

    @remote
    def avg_shared_states(self, shared_states: List[FedAvgSharedState]) -> FedAvgAveragedState:
        """Compute the weighted average of all elements returned by the train
        methods of the user-defined algorithm.
        The average is weighted by the proportion of the number of samples.

        Example:

            .. code-block:: python

                shared_states = [
                    {"weights": [3, 3, 3], "gradient": [4, 4, 4], "n_samples": 20},
                    {"weights": [6, 6, 6], "gradient": [1, 1, 1], "n_samples": 40},
                ]
                result = {"weights": [5, 5, 5], "gradient": [2, 2, 2]}

        Args:
            shared_states (typing.List[FedAvgSharedState]): The list of the
                shared_state returned by the train method of the algorithm for each organization.

        Raises:
            EmptySharedStatesError: The train method of your algorithm must return a shared_state
            TypeError: Each shared_state must contains the key **n_samples**
            TypeError: Each shared_state must contains at least one element to average
            TypeError: All the elements of shared_states must be similar (same keys)
            TypeError: All elements to average must be of type np.ndarray

        Returns:
            FedAvgAveragedState: A dict containing the weighted average of each input parameters
            without the passed key "n_samples".
        """
        if len(shared_states) == 0:
            raise EmptySharedStatesError(
                "Your shared_states is empty. Please ensure that "
                "the train method of your algorithm returns a FedAvgSharedState object."
            )

        assert all(
            [
                len(shared_state.parameters_update) == len(shared_states[0].parameters_update)
                for shared_state in shared_states
            ]
        ), "Not the same number of layers for every input parameters."

        n_all_samples = sum([state.n_samples for state in shared_states])

        averaged_states = list()
        for idx in range(len(shared_states[0].parameters_update)):
            states = list()
            for state in shared_states:
                states.append(state.parameters_update[idx] * (state.n_samples / n_all_samples))
            averaged_states.append(np.sum(states, axis=0))

        return FedAvgAveragedState(avg_parameters_update=averaged_states)

    def _perform_local_updates(
        self,
        algo: Algo,
        train_data_nodes: List[TrainDataNode],
        current_aggregation: Optional[SharedStateRef],
        round_idx: int,
        aggregation_id: str,
        additional_orgs_permissions: set,
        clean_models: bool,
    ):
        """Perform a local update (train on n mini-batches) of the models
        on each train data nodes.

        Args:
            algo (Algo): User defined algorithm: describes the model train and predict methods
            train_data_nodes (typing.List[TrainDataNode]): List of the organizations on which to perform
            local updates current_aggregation (SharedStateRef, Optional): Reference of an aggregation operation to
                be passed as input to each local training
            round_idx (int): Round number, it starts at 1.
            aggregation_id (str): Id of the aggregation node the shared state is given to.
            additional_orgs_permissions (set): Additional permissions to give to the model outputs
                after training, in order to test the model on an other organization.
            clean_models (bool): Clean the intermediary models of this round on the Substra platform.
                Set it to False if you want to download or re-use intermediary models. This causes the disk
                space to fill quickly so should be set to True unless needed.
        """

        next_local_states = []
        next_shared_states = []

        for i, node in enumerate(train_data_nodes):
            # define composite tasks (do not submit yet)
            # for each composite task give description of Algo instead of a key for an algo
            next_local_state, next_shared_state = node.update_states(
                algo.train(  # type: ignore
                    node.data_sample_keys,
                    shared_state=current_aggregation,
                    _algo_name=f"Training with {algo.__class__.__name__}",
                ),
                local_state=self._local_states[i] if self._local_states is not None else None,
                round_idx=round_idx,
                authorized_ids=set([node.organization_id]) | additional_orgs_permissions,
                aggregation_id=aggregation_id,
                clean_models=clean_models,
            )
            # keep the states in a list: one/organization
            next_local_states.append(next_local_state)
            next_shared_states.append(next_shared_state)

        self._local_states = next_local_states
        self._shared_states = next_shared_states



from pathlib import Path
from typing import Any
from typing import Dict
from typing import Type
from typing import TypedDict

import substratools as tools

from substrafl.nodes.node import InputIdentifiers
from substrafl.nodes.node import OutputIdentifiers
from substrafl.remote.serializers.pickle_serializer import PickleSerializer
from substrafl.remote.serializers.serializer import Serializer


class RemoteMethod:
    """Methods to register to Substra"""

    def __init__(
        self,
        instance,
        method_name: str,
        method_parameters: Dict,
        shared_state_serializer: Type[Serializer] = PickleSerializer,
    ):
        self.instance = instance

        self.method_name = method_name
        self.method_parameters = method_parameters

        self.shared_state_serializer = shared_state_serializer

    def load_method_inputs(self, inputs: TypedDict, outputs: TypedDict):
        """Load the different parameters needed from the inputs and outputs dictionaries
        and increment a loaded_inputs dictionary depending on the InputIdentifiers or
        OutputIdentifiers of the parameter.

        Args:
            inputs (TypedDict):  dictionary containing the paths where to load the arguments for the method.
            outputs (TypedDict):  dictionary containing the paths where to save the output for the method.

        Returns:
            TypedDict: dictionary containing the kwargs of the method to call.
        """

        loaded_inputs = {}

        instance_path = inputs.get(InputIdentifiers.local)
        if instance_path is not None:
            self.instance = self.load_instance(instance_path)

        if InputIdentifiers.models in inputs:
            models = []
            for m_path in inputs[InputIdentifiers.models]:
                models.append(self.load_model(m_path))
            loaded_inputs["shared_states"] = models

        if InputIdentifiers.datasamples in inputs:
            loaded_inputs["datasamples"] = inputs[InputIdentifiers.datasamples]

        if InputIdentifiers.shared in inputs:
            loaded_inputs["shared_state"] = (
                self.load_model(inputs[InputIdentifiers.shared]) if inputs[InputIdentifiers.shared] else None
            )

        if InputIdentifiers.predictions in inputs:
            loaded_inputs["predictions_path"] = inputs[InputIdentifiers.predictions]

        if OutputIdentifiers.predictions in outputs:
            loaded_inputs["predictions_path"] = outputs[OutputIdentifiers.predictions]

        return loaded_inputs

    def save_method_output(self, method_output: Any, outputs: TypedDict):
        """Save the method output on the path given in outputs,
        depending on the value of the OutputIdentifiers.

        Args:
            method_output (Any): return value from the called method.
            outputs (TypedDict): dictionary containing the paths where to save the output for the method.
        """

        if OutputIdentifiers.local in outputs:
            self.save_instance(outputs[OutputIdentifiers.local])

        if OutputIdentifiers.model in outputs:
            self.save_model(method_output, outputs[OutputIdentifiers.model])

        elif OutputIdentifiers.shared in outputs:
            self.save_model(method_output, outputs[OutputIdentifiers.shared])

        elif OutputIdentifiers.performance in outputs:
            tools.save_performance(method_output, outputs[OutputIdentifiers.performance])

    def generic_function(
        self,
        inputs: TypedDict,
        outputs: TypedDict,  # outputs contains a dict where keys are identifiers and values are paths on disk
        task_properties: TypedDict,
    ) -> None:
        """Generic function to be registered and executed on the Substra platform using substra-tools.

        Args:
            inputs (TypedDict): dictionary containing the paths where to load the arguments for the method.
            outputs (TypedDict): dictionary containing the paths where to save the output of the method.
            task_properties (TypedDict): Unused.
        """

        method_inputs = self.load_method_inputs(inputs, outputs)
        method_to_call = getattr(self.instance, self.method_name)

        method_inputs["_skip"] = True

        method_output = method_to_call(
            **method_inputs,
            **self.method_parameters,
        )

        self.save_method_output(method_output, outputs)

    def load_model(self, path: str) -> Any:
        """Load the model from disk

        Args:
            path (str): path to the saved model

        Returns:
            Any: loaded model
        """
        return self.shared_state_serializer.load(Path(path))

    def save_model(self, model, path: str) -> None:
        """Save the model

        Args:
            model (Any): Model to save
            path (str): Path where to save the model
        """
        self.shared_state_serializer.save(model, Path(path))

    def load_instance(self, path: str) -> Any:
        """Load the instance from disk

        Args:
            path (str): path to the saved instance

        Returns:
            Any: loaded instance
        """
        return self.instance.load(Path(path))

    def save_instance(self, path: str) -> None:
        """Save the instance

        Args:
            model (Any): Instance to save
            path (str): Path where to save the instance
        """
        self.instance.save(Path(path))

    def register_substratools_function(self):
        """Register the function that can be accessed and executed by substratools."""

        tools.register(
            function=self.generic_function,
            function_name=self.method_name,
        )



import logging
from enum import IntEnum
from pathlib import Path
from typing import Any
from typing import List
from typing import Optional

import torch

from substrafl.algorithms.pytorch import weight_manager
from substrafl.algorithms.pytorch.torch_base_algo import TorchAlgo
from substrafl.exceptions import NumUpdatesValueError
from substrafl.exceptions import ScaffoldLearningRateError
from substrafl.exceptions import TorchScaffoldAlgoParametersUpdateError
from substrafl.index_generator import BaseIndexGenerator
from substrafl.remote import remote_data
from substrafl.schemas import ScaffoldAveragedStates
from substrafl.schemas import ScaffoldSharedState
from substrafl.schemas import StrategyName

logger = logging.getLogger(__name__)


class CUpdateRule(IntEnum):
    """The rule used to update the client control variate

    Values:

        - STABLE (1): The stable rule, I in the Scaffold paper (not implemented)
        - FAST (2): The fast rule, II in the Scaffold paper
    """

    STABLE = 1
    FAST = 2


class TorchScaffoldAlgo(TorchAlgo):
    """To be inherited. Wraps the necessary operation so a torch model can be trained in the Scaffold strategy.

    The ``train`` method:

        - updates the weights of the model with the aggregated weights,
        - initializes or loads the index generator,
        - calls the :py:func:`~substrafl.algorithms.pytorch.torch_scaffold_algo.TorchScaffoldAlgo._local_train` method
          to do the local training
        - then gets the weight updates from the models and sends them to the aggregator.

    The ``predict`` method generates the predictions.

    The child class can override the
    :py:func:`~substrafl.algorithms.pytorch.torch_scaffold_algo.TorchScaffoldAlgo._local_train` and
    :py:func:`~substrafl.algorithms.pytorch.torch_scaffold_algo.TorchScaffoldAlgo._local_predict` methods,
    or other methods if necessary.

    To add a custom parameter to the ``__init__``of the class, also add it to the call to ``super().__init__```
    as shown in the example with ``my_custom_extra_parameter``. Only primitive types (str, int, ...) are supported
    for extra parameters.

    Example:

        .. code-block:: python

            class MyAlgo(TorchScaffoldAlgo):
                def __init__(
                    self,
                    my_custom_extra_parameter,
                ):
                    super().__init__(
                        model=perceptron,
                        criterion=torch.nn.MSELoss(),
                        optimizer=optimizer,
                        num_updates=100,
                        index_generator=NpIndexGenerator(
                            num_updates=10,
                            batch_size=32,
                        ),
                        dataset=MyDataset,
                        my_custom_extra_parameter=my_custom_extra_parameter,
                    )
                def _local_train(
                    self,
                    train_dataset: torch.utils.data.Dataset,
                ):
                    # Create torch dataloader
                    # ``train_dataset = self._dataset(datasamples=datasamples, is_inference=False)`` is executed
                    # prior the execution of this function
                    train_data_loader = torch.utils.data.DataLoader(train_dataset, batch_sampler=self._index_generator)

                    for x_batch, y_batch in train_data_loader:

                        # Forward pass
                        y_pred = self._model(x_batch)
                        # Compute Loss
                        loss = self._criterion(y_pred, y_batch)
                        self._optimizer.zero_grad()
                        # backward pass: compute the gradients
                        loss.backward()
                        # forward pass: update the weights.
                        self._optimizer.step()

                        # Scaffold specific: to keep between _optimizer.step() and _scheduler.step()
                        # _scheduler and Scaffold strategies are not scientifically validated, it is not
                        # recommended to use one. If one is used, _scheduler.step() must be called after
                        # _scaffold_parameters_update()
                        self._scaffold_parameters_update()
                        if self._scheduler is not None:
                            self._scheduler.step()
    """

    def __init__(
        self,
        model: torch.nn.Module,
        criterion: torch.nn.modules.loss._Loss,
        optimizer: torch.optim.Optimizer,
        index_generator: BaseIndexGenerator,
        dataset: torch.utils.data.Dataset,
        scheduler: Optional[torch.optim.lr_scheduler._LRScheduler] = None,
        with_batch_norm_parameters: bool = False,
        c_update_rule: CUpdateRule = CUpdateRule.FAST,
        seed: Optional[int] = None,
        use_gpu: bool = True,
        *args,
        **kwargs,
    ):
        """The ``__init__`` function is called at each call of the ``train`` or ``predict`` function
        For round>2, some attributes will then be overwritten by their previous states in the `load()` function,
        before the `train()` or `predict()` function is ran.

        Args:
            model (torch.nn.modules.module.Module): A torch model.
            criterion (torch.nn.modules.loss._Loss): A torch criterion (loss).
            optimizer (torch.optim.Optimizer): A torch optimizer linked to the model.
            index_generator (BaseIndexGenerator): a stateful index generator.
                Must inherit from BaseIndexGenerator. The __next__ method shall return a python object (batch_index)
                which is used for selecting each batch from the output of the get_data method of the opener
                during training in this way: ``x[batch_index], y[batch_index]``.
                If overridden, the generator class must be defined either as part of a package or in a different file
                than the one from which the ``execute_experiment`` function is called.
                This generator is used as stateful ``batch_sampler`` of the data loader created from the given
                ``dataset``
            dataset (torch.utils.data.Dataset): an instantiable dataset class whose ``__init__`` arguments are
                ``x``, ``y`` and ``is_inference``. The torch datasets used for both training and inference will be
                instantiate from it prior to the ``_local_train`` execution and within the ``predict`` method.
                The ``__getitem__`` methods of those generated datasets must return both ``x`` (training data) and y
                (target values) when ``is_inference`` is set to ``False`` and only ``x`` (testing data) when
                ``is_inference`` is set to True.
                This behavior can be changed by re-writing the `_local_train` or `predict` methods.
            scheduler (torch.optim.lr_scheduler._LRScheduler, Optional): A torch scheduler that will be called at every
                batch. If None, no scheduler will be used. Defaults to None.
            with_batch_norm_parameters (bool): Whether to include the batch norm layer parameters in the fed avg
                strategy. Defaults to False.
            c_update_rule (CUpdateRule): The rule used to update the
                client control variate.
                Defaults to CUpdateRule.FAST.
            seed (typing.Optional[int]): Seed set at the algo initialization on each organization. Defaults to None.
            use_gpu (bool): Whether to use the GPUs if they are available. Defaults to True.
        Raises:
            :ref:`~substrafl.exceptions.NumUpdatesValueError`: If `num_updates` is inferior or equal to zero.
        """
        super().__init__(
            model=model,
            criterion=criterion,
            optimizer=optimizer,
            index_generator=index_generator,
            dataset=dataset,
            scheduler=scheduler,
            use_gpu=use_gpu,
            seed=seed,
            *args,
            **kwargs,
        )

        if self._index_generator.num_updates <= 0:
            raise NumUpdatesValueError("Num_updates needs to be superior to 0 for TorchScaffoldAlgo.")

        if not isinstance(self._optimizer, torch.optim.SGD):
            logger.warning("The only optimizer theoretically guaranteed to work with the Scaffold strategy is SGD.")

        self._lr_warnings()

        self._with_batch_norm_parameters = with_batch_norm_parameters
        self._c_update_rule = CUpdateRule(c_update_rule)
        # ci in the paper
        self._client_control_variate: List[torch.Tensor] = None
        # c in the paper
        self._server_control_variate: List[torch.Tensor] = None
        # the lr used by optimizer.step()
        self._current_lr: float = None
        # the delta_variate used in _scaffold_parameters_update()
        self._delta_variate: List[torch.Tensor] = None
        # Private attributes to monitor the use of _scaffold_parameters_update()
        # Must be set to 0 at the beginning of each task execution
        self._scaffold_parameters_update_num_call = 0

    @property
    def strategies(self) -> List[StrategyName]:
        """List of compatible strategies

        Returns:
            typing.List: typing.List[StrategyName]
        """
        return [StrategyName.SCAFFOLD]

    def _lr_warnings(self, learning_rates=None):
        """Scientific warnings about the use of different learning rates during training."""
        # Retrieve all lr defined per layer
        if learning_rates is None:
            learning_rates = set([param_group.get("lr") for param_group in self._optimizer.param_groups])
            learning_rates.discard(0)

        if self._scheduler:
            logger.warning(
                """Scaffold strategies and the use of a scheduler has not been scientifically validated."""
                """\nIf used, self._scheduler.step() must be called after self._scaffold_parameters_update()."""
            )

        if len(learning_rates) == 0:
            # Torch needs a learning rates for each layer to set the optimizer so this case should never happen.
            # Keeping it for consistency
            raise ScaffoldLearningRateError(
                "When using the  Torch Scaffold algo layer, a learning rate must "
                "be passed to for all group of layers in the optimizer and one of them shall be "
                "strictly positive."
            )

        elif len(learning_rates) > 1:
            logger.warning(
                "Different learning rates where found from the optimizer: %s. "
                "The aggregation operation of the Scaffold strategy, requires a unique learning rate. "
                "Hence the smallest one will be used.",
                str(learning_rates),
            )

    def _update_current_lr(self):
        """Update the `self._current_lr` attributes from the optimizer. It will be updated from the scheduler if
        one is used even if this is not recommended as the use of a scheduler with the Scaffold strategy has not
        been scientifically validated.
        Different use cases: https://discuss.pytorch.org/t/get-current-lr-of-optimizer-with-adaptive-lr/24851/6
        """

        if self._scheduler and self._scheduler.get_last_lr():
            learning_rates = set(self._scheduler.get_last_lr())
        else:
            # Retrieve all lr defined per layer
            learning_rates = set([param_group.get("lr") for param_group in self._optimizer.param_groups])

        # If 0 is set as a learning rate for some layers, the weights and biases won't be impacted by the training
        # This behavior is accepted
        learning_rates.discard(0)

        self._lr_warnings(learning_rates=learning_rates)

        self._current_lr = float(min(learning_rates))

    def _scaffold_parameters_update(self):
        """Must be called for each update after the optimizer.step() operation."""
        # Adding control variates on weights times learning rate
        # Scaffold paper's Algo step 10.2 :  yi = last_yi - lr * ( - ci + c) = last_yi + lr * ( ci - c)
        # <=> model = last_model + lr * delta_variate
        self._update_current_lr()
        self._scaffold_parameters_update_num_call += 1
        weight_manager.increment_parameters(
            model=self._model,
            updates=self._delta_variate,
            with_batch_norm_parameters=self._with_batch_norm_parameters,
            updates_multiplier=self._current_lr,
        )

    def _local_train(
        self,
        train_dataset: torch.utils.data.Dataset,
    ):
        """Local train method. Contains the local training loop.

        Train the model on ``num_updates`` minibatches, using the ``self._index_generator generator`` as batch sampler
        for the torch dataset.

        Args:
            train_dataset (torch.utils.data.Dataset): train_dataset build from the x and y returned by the opener.

        Important:

            You must use ``next(self._index_generator)`` at each minibatch,
            to ensure that you are using the batches are correct between 2 rounds
            of the federated learning strategy.

        Example:

            .. code-block:: python

                # Create torch dataloader
                train_data_loader = torch.utils.data.DataLoader(train_dataset, batch_sampler=self._index_generator)

                for x_batch, y_batch in train_data_loader:

                    # Forward pass
                    y_pred = self._model(x_batch)

                    # Compute Loss
                    loss = self._criterion(y_pred, y_batch)
                    self._optimizer.zero_grad()
                    loss.backward()
                    self._optimizer.step()

                    # Scaffold specific function to call between self._optimizer.step() and self._scheduler.step()
                    self._scaffold_parameters_update()

                    if self._scheduler is not None:
                        self._scheduler.step()
        """

        # Create torch dataloader
        train_data_loader = torch.utils.data.DataLoader(train_dataset, batch_sampler=self._index_generator)

        for x_batch, y_batch in train_data_loader:

            x_batch = x_batch.to(self._device)
            y_batch = y_batch.to(self._device)

            # Forward pass
            y_pred = self._model(x_batch)
            # Compute Loss
            loss = self._criterion(y_pred, y_batch)
            self._optimizer.zero_grad()
            # backward pass: compute the gradients
            loss.backward()
            # forward pass: update the weights.
            # if optimizer=SGD, yi = last_yi - lr * grads,
            # with the same lr as _get_current_lr() (used in step 10.2)
            self._optimizer.step()

            # Scaffold specific: to call between self._optimizer.step() and self._scheduler.step()
            self._scaffold_parameters_update()

            if self._scheduler is not None:
                self._scheduler.step()

    @remote_data
    def train(
        self,
        datasamples: Any,
        shared_state: Optional[ScaffoldAveragedStates] = None,  # Set to None per default for clarity reason as
        #  the decorator will do it if the arg shared_state is not passed.
    ) -> ScaffoldSharedState:
        """Train method of the Scaffold strategy implemented with torch. This method will execute the following
        operations:

            * instantiates the provided (or default) batch indexer
            * if a shared state is passed, set the parameters of the model to the provided shared state
            * train the model for n_updates
            * compute the weight update and control variate update

        Args:
            datasamples (typing.Any): Input data returned by the ``get_data`` method from the opener.
            shared_state (typing.Optional[ScaffoldAveragedStates]): Shared state sent by the aggregate_organization
                (returned by the func strategies.scaffold.avg_shared_states)
                Defaults to None.


        Returns:
            ScaffoldSharedState: the shared states of the Algo
        """

        # Create torch dataset
        train_dataset = self._dataset(datasamples, is_inference=False)

        if shared_state is None:  # first round
            # Instantiate the index_generator
            assert self._index_generator.n_samples is None
            self._index_generator.n_samples = len(train_dataset)

            # client_control_variate = zeros matrix with the shape of the model weights
            assert self._client_control_variate is None
            self._client_control_variate = weight_manager.zeros_like_parameters(
                self.model,
                with_batch_norm_parameters=self._with_batch_norm_parameters,
                device=self._device,
            )
            # we initialize the server_control_variate (c in the paper) here so we don't have to do
            # an initialization round
            assert self._server_control_variate is None
            self._server_control_variate = weight_manager.zeros_like_parameters(
                self.model,
                with_batch_norm_parameters=self._with_batch_norm_parameters,
                device=self._device,
            )
        else:  # round>1
            # These should have been loaded by the load() function
            assert self._client_control_variate is not None
            assert self._index_generator.n_samples is not None

            # The shared states is the average of the difference of the parameters_update for all organizations
            # Hence we need to add it to the previous local state parameters
            # Scaffold paper's Algo step 17: model = model + aggregation_lr * parameters_update
            # here shared_state.avg_parameters_update is already aggregation_lr * parameters_update,
            # cf strategies.scaffold.avg_shared_states
            avg_parameters_update = [torch.from_numpy(x).to(self._device) for x in shared_state.avg_parameters_update]
            weight_manager.increment_parameters(
                model=self._model,
                updates=avg_parameters_update,
                with_batch_norm_parameters=self._with_batch_norm_parameters,
            )
            # get the server_control_variate from the aggregator
            self._server_control_variate = [
                torch.from_numpy(t).to(self._device) for t in shared_state.server_control_variate
            ]

        self._index_generator.reset_counter()

        # save original parameters to compute the weight updates and reset the model parameters after
        # the model is trained
        original_parameters = weight_manager.get_parameters(
            model=self._model, with_batch_norm_parameters=self._with_batch_norm_parameters
        )

        # compute delta_variate = ci-c for Scaffold paper's Algo step 10.2 (self._scaffold_parameters_update())
        self._delta_variate = weight_manager.subtract_parameters(
            parameters=self._client_control_variate,
            parameters_to_subtract=self._server_control_variate,
        )

        # Train mode for torch model
        self._model.train()

        # Train the model
        self._local_train(train_dataset)

        self._index_generator.check_num_updates()

        if self._scaffold_parameters_update_num_call != self._index_generator._num_updates:
            raise TorchScaffoldAlgoParametersUpdateError(
                f"`_scaffold_parameters_update` method has been called {self._scaffold_parameters_update_num_call} "
                f"time(s) but num_updates is set to {self._index_generator._num_updates}. Please check within your "
                "`_local_train` function that `_scaffold_parameters_update` is called at each update (each time "
                "self._model(data) is called) after the `self.optimizer.step()` call."
            )

        self._model.eval()

        # Scaffold paper's Algo step 12+13.1: compute parameters_update = (yi-x)
        parameters_update = weight_manager.subtract_parameters(
            parameters=weight_manager.get_parameters(
                model=self._model,
                with_batch_norm_parameters=self._with_batch_norm_parameters,
            ),
            parameters_to_subtract=original_parameters,
        )

        if self._c_update_rule == CUpdateRule.FAST:
            # right_multiplier = -1 / (lr*num_updates)
            right_multiplier = -1.0 / (self._current_lr * self._index_generator.num_updates)
            # Scaffold paper's Algo step 12+13.2: control_variate_update = -c - parameters_update / (lr*num_updates)
            control_variate_update = weight_manager.weighted_sum_parameters(
                parameters_list=[self._server_control_variate, parameters_update],
                coefficient_list=[-1.0, right_multiplier],
            )
        else:
            raise NotImplementedError("rule 1 not implemented")

        # Scaffold paper's Algo step 14: ci = ci + control_variate_update
        self._client_control_variate = weight_manager.add_parameters(
            parameters=self._client_control_variate,
            parameters_to_add=control_variate_update,
        )

        # Re set to the previous state
        weight_manager.set_parameters(
            model=self._model,
            parameters=original_parameters,
            with_batch_norm_parameters=self._with_batch_norm_parameters,
        )

        # Scaffold paper's Algo step 13: return model_weight_update & control_variate_update
        return_dict = ScaffoldSharedState(
            parameters_update=[w.cpu().detach().numpy() for w in parameters_update],
            control_variate_update=[c.cpu().detach().numpy() for c in control_variate_update],
            server_control_variate=[s.cpu().detach().numpy() for s in self._server_control_variate],
            n_samples=len(train_dataset),
        )
        return return_dict

    def _get_state_to_save(self) -> dict:
        """Get the local state to save, the only strategy-specific variable
        to save is the ``client_control_variate``.

        Returns:
            dict: checkpoint
        """
        local_state = super()._get_state_to_save()
        local_state.update(
            {
                "client_control_variate": self._client_control_variate,
            }
        )
        return local_state

    def _update_from_checkpoint(self, path: Path) -> dict:
        """Load the local state from the checkpoint.

        Args:
            path (pathlib.Path): path where the checkpoint is saved

        Returns:
            dict: checkpoint
        """
        checkpoint = super()._update_from_checkpoint(path=path)
        self._client_control_variate = checkpoint.pop("client_control_variate")
        return checkpoint

    def summary(self):
        """Summary of the class to be exposed in the experiment summary file

        Returns:
            dict: a json-serializable dict with the attributes the user wants to store
        """
        summary = super().summary()
        summary.update(
            {
                "with_batch_norm_parameters": str(self._with_batch_norm_parameters),
            }
        )
        return summary



import inspect
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Optional
from typing import Type

import cloudpickle

from substrafl.remote.substratools_methods import RemoteMethod


class RemoteStruct:
    """Contains the wrapped user code and the necessary functions
    to transform it into a Substra asset to execute on the platform.

    Args:
        cls (Type): The remote struct type (e.g. Algorithm, dataset)
        cls_parameters (str): The class parameters serialized into json string.
            E.g.: use ``json.dumps({"args": [], "kwargs": kwargs})``
        remote_cls (str): The name of the class used remotely
        remote_cls_parameters (str): The remote class parameters serialized into json string.
            E.g.: use ``json.dumps({"args": [], "kwargs": kwargs})``
        algo_name(str, Optional): opportunity to set a custom algo name.
            If None, set to "{method_name}_{class_name}"
    """

    def __init__(
        self,
        cls: Type,
        cls_args: list,
        cls_kwargs: dict,
        remote_cls: Type[RemoteMethod],
        method_name: str,
        method_parameters: dict,
        algo_name: Optional[str],
    ):
        """
        Args:
            cls (Type): Locally defined class
            cls_args (list): Arguments (args) to instantiate the class
            cls_kwargs (dict): Arguments (kwargs) to instantiate the class
            remote_cls (Type[RemoteMethod]): Remote class to create from the user code
            method_name (str): Name of the method from the local class to execute
            method_parameters (dict): Parameters to pass to the method
            algo_name(str, Optional): opportunity to set a custom algo name.
                If None, set to "{method_name}_{class_name}"
        """
        self._cls = cls
        self._cls_args = cls_args
        self._cls_kwargs = cls_kwargs
        self._remote_cls = remote_cls
        self._method_name = method_name
        self._method_parameters = method_parameters
        self._algo_name = algo_name or (self._method_name + "_" + self._cls.__name__)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, RemoteStruct):
            return NotImplemented
        return (
            self._cls == other._cls
            and self._cls_args == other._cls_args
            and self._cls_kwargs == other._cls_kwargs
            and self._remote_cls == other._remote_cls
            and self._method_name == other._method_name
            and self._method_parameters == other._method_parameters
        )

    def __hash__(self):
        return hash(
            (
                self._cls,
                frozenset(self._cls_args),
                frozenset(self._cls_kwargs),
                self._remote_cls,
                self._method_name,
                frozenset(self._method_parameters),
            )
        )

    @property
    def algo_name(self):
        return self._algo_name

    @classmethod
    def load(cls, src: Path) -> "RemoteStruct":
        """Load the remote struct from the src
        directory.

        Args:
            src (pathlib.Path): Path to the directory where the remote struct has been saved.
        """
        cls_cloudpickle_path = src / "cls_cloudpickle"
        with cls_cloudpickle_path.open("rb") as f:
            instance = cloudpickle.load(f)
        return instance

    def save(self, dest: Path):
        """Save the instance to the dest directory using
        cloudpickle.

        Args:
            dest (pathlib.Path): directory where to save the remote struct
        """
        cloudpickle_path = dest / "cls_cloudpickle"
        with cloudpickle_path.open("wb") as f:
            cloudpickle.dump(self, f)

    def get_instance(self) -> Any:
        """Get the class instance.

        Returns:
            typing.Any: Instance of the saved class
        """
        return self._cls(*self._cls_args, **self._cls_kwargs)

    def get_remote_instance(self) -> RemoteMethod:
        """Get the remote class (ie Substra algo) instance.

        Returns:
            RemoteMethod: instance of the remote Substra class
        """
        return self._remote_cls(
            self.get_instance(),
            method_name=self._method_name,
            method_parameters=self._method_parameters,
        )

    def get_cls_file_path(self) -> Path:
        """Get the path to the file where the cls attribute is defined.

        Returns:
            pathlib.Path: path to the file where the cls is defined.
        """
        try:
            algo_file_path = Path(inspect.getfile(self._cls)).resolve().parent
        except (TypeError, OSError):
            # In a notebook, we get the TypeError (OSError on python 3.10): <class '__main__.MyAlgo'> is a built-in
            # class.
            # To fix it, we use the cwd of the notebook and assume local dependencies are there
            algo_file_path = Path.cwd().resolve()
        return algo_file_path

    def summary(self) -> Dict[str, str]:
        """Get a summary of what the remote struct represents.

        Returns:
            typing.Dict[str, str]: description
        """
        return {
            "type": self._cls.__name__,
            "method_name": self._method_name,
        }



from substrafl.index_generator.base import BaseIndexGenerator
from substrafl.index_generator.np_index_generator import NpIndexGenerator

__all__ = [
    "BaseIndexGenerator",
    "NpIndexGenerator",
]



from torch import __version__ as torch_version

from substrafl.algorithms.pytorch.torch_fed_avg_algo import TorchFedAvgAlgo
from substrafl.algorithms.pytorch.torch_newton_raphson_algo import TorchNewtonRaphsonAlgo
from substrafl.algorithms.pytorch.torch_scaffold_algo import TorchScaffoldAlgo
from substrafl.algorithms.pytorch.torch_single_organization_algo import TorchSingleOrganizationAlgo
from substrafl.exceptions import UnsupportedPytorchVersionError

if torch_version == "1.12.0":
    raise UnsupportedPytorchVersionError(
        "Please use an other pytorch version. There is a regression bug in torch 1.12.0, that impacts optimizers that "
        "have been pickled and unpickled. "
        "This bug occurs for Adam optimizer for example (but not for SGD). Here is a link to one issue covering it: "
        "https://github.com/pytorch/pytorch/pull/80345"
    )


__all__ = ["TorchFedAvgAlgo", "TorchSingleOrganizationAlgo", "TorchScaffoldAlgo", "TorchNewtonRaphsonAlgo"]



import math
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np
import torch

from substrafl.algorithms.pytorch import weight_manager
from substrafl.algorithms.pytorch.torch_base_algo import TorchAlgo
from substrafl.exceptions import CriterionReductionError
from substrafl.exceptions import NegativeHessianMatrixError
from substrafl.index_generator import NpIndexGenerator
from substrafl.remote import remote_data
from substrafl.schemas import NewtonRaphsonAveragedStates
from substrafl.schemas import NewtonRaphsonSharedState
from substrafl.schemas import StrategyName


class TorchNewtonRaphsonAlgo(TorchAlgo):
    """To be inherited. Wraps the necessary operation so a torch model can be trained in the Newton-Raphson strategy.

    The ``train`` method:

        - updates the weights of the model with the calculated weight updates
        - creates and initializes the index generator with the given batch size
        - calls the
          :py:func:`~substrafl.algorithms.pytorch.torch_newton_raphson_algo.TorchNewtonRaphsonAlgo._local_train`
          method to compute the local gradients and Hessian and sends them to the aggregator.
        - a L2 regularization can be applied to the loss by settings ``l2_coeff`` different to zero (default value). L2
          regularization adds numerical stability when inverting the hessian.

    The child class can overwrite
    :py:func:`~substrafl.algorithms.pytorch.torch_newton_raphson_algo.TorchNewtonRaphsonAlgo._local_train`
    and :py:func:`~substrafl.algorithms.pytorch.torch_newton_raphson_algo.TorchNewtonRaphsonAlgo._local_predict`,
    or other methods if necessary.

    To add a custom parameter to the ``__init__`` of the class, also add it to the call to ``super().__init__``.
    """

    def __init__(
        self,
        model: torch.nn.Module,
        criterion: torch.nn.modules.loss._Loss,
        batch_size: Optional[int],
        dataset: torch.utils.data.Dataset,
        l2_coeff: float = 0,
        with_batch_norm_parameters: bool = False,
        seed: Optional[int] = None,
        use_gpu: bool = True,
        *args,
        **kwargs,
    ):
        """The ``__init__`` function is called at each call of the ``train`` or ``predict`` function.

        For ``round>=2``, some attributes will then be overwritten by their previous states in the
        :py:func:`~substrafl.algorithms.pytorch.torch_base_algo.TorchAlgo.load` function, before the
        ``train`` or ``predict`` function is ran.

        ``TorchNewtonRaphsonAlgo`` computes its :py:class:`~substrafl.schemas.NewtonRaphsonSharedState`
        (gradients and Hessian matrix) on all the samples of the dataset. Data might be split into mini-batches to
        prevent loading too much data at once.

        Args:
            model (torch.nn.modules.module.Module): A torch model.
            criterion (torch.nn.modules.loss._Loss): A torch criterion (loss).
            batch_size (int): The size of the batch. If set to None it will be set to the number of samples in the
                dataset. Note that dividing the data to batches is done only to avoid the memory issues. The weights
                are updated only at the end of the epoch.
            dataset (torch.utils.data.Dataset): an instantiable dataset class whose ``__init__`` arguments are
                ``x``, ``y`` and ``is_inference``. The torch datasets used for both training and inference will be
                instantiate from it prior to the ``_local_train`` execution and within the ``predict`` method.
                The ``__getitem__`` methods of those generated datasets must return both ``x`` (training data) and y
                (target values) when ``is_inference`` is set to ``False`` and only ``x`` (testing data) when
                ``is_inference`` is set to True.
                This behavior can be changed by re-writing the `_local_train` or `predict` methods.
            l2_coeff (float): L2 regularization coefficient. The larger l2_coeff is, the better the stability of the
                hessian matrix will be, however the convergence might be slower. Defaults to 0.
            with_batch_norm_parameters (bool): Whether to include the batch norm layer parameters in the Newton-Raphson
                strategy. Defaults to False.
            seed (typing.Optional[int]): Seed set at the algo initialization on each organization. Defaults to None.
            use_gpu (bool): Whether to use the GPUs if they are available. Defaults to True.
        """
        assert "optimizer" not in kwargs, "Newton Raphson strategy does not uses optimizers"

        super().__init__(
            model=model,
            criterion=criterion,
            optimizer=None,
            index_generator=None,
            dataset=dataset,
            use_gpu=use_gpu,
            seed=seed,
            *args,
            **kwargs,
        )
        self._with_batch_norm_parameters = with_batch_norm_parameters
        self._l2_coeff = l2_coeff
        self._batch_size = batch_size

        if self._criterion.reduction != "mean":
            raise CriterionReductionError(
                "The criterion reduction must be set to 'mean' to use the Newton-Raphson strategy"
            )

        # initialized and used only in the train method
        self._final_gradients = None
        self._final_hessian = None
        self._n_samples_done = None

    @property
    def strategies(self) -> List[StrategyName]:
        """List of compatible strategies

        Returns:
            typing.List: typing.List[StrategyName]
        """
        return [StrategyName.NEWTON_RAPHSON]

    def _l2_reg(self) -> torch.Tensor:
        """Compute the l2 regularization regarding the model parameters.

        Returns:
            torch.Tensor: the updated loss with the l2 regularization included.
        """
        # L2 regularization
        l2_reg = 0
        for param in self.model.parameters():
            l2_reg += self._l2_coeff * torch.sum(param**2) / 2
        return l2_reg

    def _initialize_gradients_and_hessian(self):
        """Initializes the gradients and hessian matrices"""
        number_of_trainable_params = sum(p.numel() for p in self._model.parameters() if p.requires_grad)

        n_samples_done = 0

        final_gradients = [torch.zeros_like(p).numpy() for p in self._model.parameters()]
        final_hessian = np.zeros([number_of_trainable_params, number_of_trainable_params])

        return final_gradients, final_hessian, n_samples_done

    def _update_gradients_and_hessian(self, loss: torch.Tensor, current_batch_size: int):
        """Updates the gradients and hessian matrices.

        Args:
            loss (torch.Tensor): the loss to compute the gradients and hessian from.
            current_batch_size (int): The length of the batch used to compute the given loss.
        """

        gradients, hessian = self._compute_gradients_and_hessian(loss)

        self._n_samples_done += current_batch_size

        batch_coefficient = current_batch_size / self._index_generator.n_samples

        self._final_hessian += hessian.cpu().detach().numpy() * batch_coefficient
        self._final_gradients = [
            sum(final_grad, grad.cpu().detach().numpy() * batch_coefficient)
            for final_grad, grad in zip(self._final_gradients, gradients)
        ]

    def _instantiate_index_generator(self, n_samples):
        if self._batch_size is None:
            # If batch_size is None, it is set to the number of samples in the dataset by the index generator
            num_updates = 1
        else:
            num_updates = math.ceil(float(n_samples) / self._batch_size)

        index_generator = NpIndexGenerator(batch_size=self._batch_size, num_updates=num_updates, drop_last=False)
        index_generator.n_samples = n_samples
        return index_generator

    def _local_train(
        self,
        train_dataset: torch.utils.data.Dataset,
    ):
        """Local train method. Contains the local training loop.

        Train the model on ``num_updates`` minibatches, using the ``self._index_generator generator`` as batch sampler
        for the torch dataset.

        Args:
            train_dataset (torch.utils.data.Dataset): train_dataset build from the x and y returned by the opener.

        Important:

            You must use ``next(self._index_generator)`` at each minibatch,
            to ensure that you are using the batches are correct between 2 rounds
            of the Newton Raphson strategy.

        Important:

            Call the function ``self._update_gradients_and_hessian(loss, current_batch_size)`` after computing the loss
            and the current_batch_size.

        Example:

            .. code-block:: python

                # As the parameters of the model don't change during the loop, the l2 regularization is constant and
                # can be calculated only once for all the batches.
                l2_reg = self._l2_reg()

                # Create torch dataloader
                train_data_loader = torch.utils.data.DataLoader(train_dataset, batch_sampler=self._index_generator)

                for x_batch, y_batch in train_data_loader:

                    # Forward pass
                    y_pred = self._model(x_batch)

                    # Compute Loss
                    loss = self._criterion(y_pred, y_batch)

                    # L2 regularization
                    loss += l2_reg

                    current_batch_size = len(x_batch)

                    # NEWTON RAPHSON specific function, to call after computing the loss and the current_batch_size.

                    self._update_gradients_and_hessian(loss, current_batch_size)
        """

        # As the parameters of the model don't change during the loop, the l2 regularization is constant and can be
        # calculated only once for all the batches.
        l2_reg = self._l2_reg()

        # Create torch dataloader
        train_data_loader = torch.utils.data.DataLoader(train_dataset, batch_sampler=self._index_generator)

        for x_batch, y_batch in train_data_loader:

            x_batch = x_batch.to(self._device)
            y_batch = y_batch.to(self._device)

            # Forward pass
            y_pred = self._model(x_batch)

            # Compute Loss
            loss = self._criterion(y_pred, y_batch)

            # L2 regularization
            loss += l2_reg

            current_batch_size = len(x_batch)

            self._update_gradients_and_hessian(loss, current_batch_size)

    def _local_predict(self, predict_dataset: torch.utils.data.Dataset, predictions_path):
        """Executes the following operations:

            * Create the torch dataloader using the batch size given at the ``__init__`` of the class
            * Sets the model to `eval` mode
            * Returns the predictions

        Args:
            predict_dataset (torch.utils.data.Dataset): predict_dataset build from the x returned by the opener.
        """
        predict_loader = torch.utils.data.DataLoader(predict_dataset, batch_size=self._batch_size)

        self._model.eval()

        predictions = torch.Tensor([])
        with torch.inference_mode():
            for x in predict_loader:
                x = x.to(self._device)
                predictions = torch.cat((predictions, self._model(x)), 0)

        predictions = predictions.cpu().detach()

        self._save_predictions(predictions, predictions_path)

    @remote_data
    def train(
        self,
        datasamples: Any,
        # Set shared_state to None per default for clarity reason as
        # the decorator will do it if the arg shared_state is not passed.
        shared_state: Optional[NewtonRaphsonAveragedStates] = None,
    ) -> NewtonRaphsonSharedState:
        """Train method of the Newton Raphson strategy implemented with torch. This method will execute the following
        operations:

            * creates and initializes the index generator
            * if a shared state is passed, set the weights of the model to the provided shared state weights
            * initializes hessians and gradient
            * calls the
                :py:func:`~substrafl.algorithms.pytorch.torch_newton_raphson_algo.TorchNewtonRaphsonAlgo._local_train`
                method to compute the local gradients and Hessian and sends them to the aggregator.
            * a L2 regularization can be applied to the loss by settings `l2_coeff` different to zero (default value)

        Args:
            datasamples (typing.Any): Input data returned by the ``get_data`` method from the opener.
            shared_state (NewtonRaphsonAveragedStates, Optional): Dict containing torch parameters that
                will be set to the model. Defaults to None.

        Returns:
            NewtonRaphsonSharedState: local gradients, local Hessian and the number of samples they were computed from.

        Raises:
            NegativeHessianMatrixError: Hessian matrix must be positive semi-definite to correspond to a convex problem.
        """

        # Create torch dataset
        train_dataset = self._dataset(datasamples, is_inference=False)

        if shared_state is None:
            # Instantiate the index_generator
            n_samples = len(train_dataset)

            self._index_generator = self._instantiate_index_generator(n_samples)

        else:
            assert self._index_generator.n_samples is not None

            # The shared states are the model parameter updates.
            # Hence we need to add it to the previous local state parameters.
            # unflatten_parameters_update = self._unflatten_tensor(
            #     shared_state.parameters_update, [p for p in self.model.parameters()]
            # )
            parameter_updates = [torch.from_numpy(x).to(self._device) for x in shared_state.parameters_update]
            weight_manager.increment_parameters(
                model=self._model,
                updates=parameter_updates,
                with_batch_norm_parameters=self._with_batch_norm_parameters,
            )
        self._index_generator.reset_counter()

        # Train mode for torch model
        self._model.train()

        self._final_gradients, self._final_hessian, self._n_samples_done = self._initialize_gradients_and_hessian()

        self._local_train(train_dataset)

        # Newton Raphson strategy must go through all the samples before each next update.
        assert self._index_generator.n_samples == self._n_samples_done

        eigenvalues = np.linalg.eig(self._final_hessian)[0].real
        if not (eigenvalues >= 0).all():
            raise NegativeHessianMatrixError(
                "Hessian matrix is not positive semi-definite, either the problem is not convex or due to numerical"
                " instability. It is advised to try to increase the l2_coeff. "
                f"Calculated eigenvalues are {eigenvalues.tolist()} and considered l2_coeff is {self._l2_coeff}"
            )
        self._index_generator.check_num_updates()

        return NewtonRaphsonSharedState(
            n_samples=self._index_generator.n_samples,
            hessian=self._final_hessian,
            gradients=self._final_gradients,
        )

    def _jacobian(self, tensor_y: torch.Tensor, create_graph: bool = False) -> torch.Tensor:
        """Compute the Jacobian for each  the given tensor_y regarding the model parameters.

        Args:
            tensor_y (torch.Tensor): _description_
            create_graph (bool, optional): Create the graph to compute higher order derivative. Defaults to False.

        Returns:
            torch.Tensor: Jacobian.
        """
        jacobian = []
        flat_y = torch.cat([t.reshape(-1) for t in tensor_y])
        for y in flat_y:
            for param in self._model.parameters():
                if param.requires_grad:
                    (gradient,) = torch.autograd.grad(y, param, retain_graph=True, create_graph=create_graph)
                    jacobian.append(gradient)

        return jacobian

    def _hessian_shape(self, second_order_derivative: List[torch.Tensor]) -> torch.Tensor:
        """Reshape from the second order derivative to obtain the Hessian matrix.

        Args:
            second_order_derivative (List[torch.Tensor]): second order derivative of a tensor regarding the
            registered parameters

        Returns:
            torch.Tensor: Hessian matrix
        """
        hessian = torch.cat([t.reshape(-1) for t in second_order_derivative])
        return hessian.reshape(self._final_hessian.shape)

    def _compute_gradients_and_hessian(self, loss: torch.Tensor) -> Tuple[torch.Tensor]:
        """The compute_gradients_and_hessian function compute the gradients and the Hessian matrix of the parameters
        regarding the given loss, and outputs them.

        Args:
            loss (torch.Tensor): the loss to compute the gradients and Hessian on.

        Returns:
            torch.Tensor: the computed gradients of the parameters regarding the loss, flattened into a 1d torch
            Tensor.
            torch.Tensor: the computed Hessian matrix of the parameters regarding the loss.
        """

        gradients = self._jacobian(loss[None], create_graph=True)
        second_order_derivative = self._jacobian(gradients)

        hessian = self._hessian_shape(second_order_derivative)

        return gradients, hessian

    def summary(self):
        """Summary of the class to be exposed in the experiment summary file

        Returns:
            dict: a json-serializable dict with the attributes the user wants to store
        """
        summary = super().summary()
        summary.update(
            {
                "with_batch_norm_parameters": str(self._with_batch_norm_parameters),
            }
        )
        return summary
